import copy

import pytest

from ffun.core.tests.helpers import TableSizeDelta, TableSizeNotChanged
from ffun.domain.domain import new_entry_id
from ffun.domain.entities import EntryId
from ffun.library.entities import Entry
from ffun.ontology.domain import (
    _inplace_filter_out_entry_tags,
    apply_tags_to_entry,
    get_ids_by_uids,
    get_tags_ids_for_entries,
    prepare_tags_for_entries,
)
from ffun.ontology.entities import ProcessorTag
from ffun.tags import converters


class TestApplyTagsToEntry:

    @pytest.mark.asyncio
    async def test_apply_no_tags(self, cataloged_entry: Entry, fake_processor_id: int) -> None:
        async with (
            TableSizeNotChanged("o_tags_properties"),
            TableSizeNotChanged("o_relations"),
            TableSizeNotChanged("o_relations_processors"),
        ):
            await apply_tags_to_entry(cataloged_entry.id, fake_processor_id, tags=[])

        loaded_tags = await get_tags_ids_for_entries([cataloged_entry.id])

        assert loaded_tags == {}

    @pytest.mark.asyncio
    async def test_apply_tags(
        self,
        cataloged_entry: Entry,
        fake_processor_id: int,
        three_processor_tags: tuple[ProcessorTag, ProcessorTag, ProcessorTag],
    ) -> None:

        three_processor_tags[1].link = "https://example.com"

        raw_to_uids = {tag.raw_uid: converters.normalize(tag.raw_uid) for tag in three_processor_tags}
        uids_to_ids = await get_ids_by_uids(raw_to_uids.values())

        async with (
            TableSizeDelta("o_tags_properties", delta=1),
            TableSizeDelta("o_relations", delta=3),
            TableSizeDelta("o_relations_processors", delta=3),
        ):
            await apply_tags_to_entry(cataloged_entry.id, fake_processor_id, three_processor_tags)

        loaded_tags = await get_tags_ids_for_entries([cataloged_entry.id])

        assert loaded_tags == {cataloged_entry.id: set(uids_to_ids.values())}

    @pytest.mark.asyncio
    async def test_duplicated_tags(
        self,
        cataloged_entry: Entry,
        fake_processor_id: int,
        three_processor_tags: tuple[ProcessorTag, ProcessorTag, ProcessorTag],
    ) -> None:

        three_processor_tags[0].raw_uid = three_processor_tags[0].raw_uid.lower()
        three_processor_tags[0].link = "https://example.com?x"

        three_processor_tags[1].link = "https://example.com?y"

        three_processor_tags[2].raw_uid = three_processor_tags[0].raw_uid.upper()
        three_processor_tags[2].link = "https://example.com?z"

        raw_to_uids = {tag.raw_uid: converters.normalize(tag.raw_uid) for tag in three_processor_tags}
        uids_to_ids = await get_ids_by_uids(raw_to_uids.values())

        async with (
            TableSizeDelta("o_tags_properties", delta=2),
            TableSizeDelta("o_relations", delta=2),
            TableSizeDelta("o_relations_processors", delta=2),
        ):
            await apply_tags_to_entry(cataloged_entry.id, fake_processor_id, three_processor_tags)

        loaded_tags = await get_tags_ids_for_entries([cataloged_entry.id])

        assert loaded_tags == {
            cataloged_entry.id: {
                uids_to_ids[raw_to_uids[three_processor_tags[0].raw_uid]],
                uids_to_ids[raw_to_uids[three_processor_tags[1].raw_uid]],
            }
        }


def _no_type_inplace_filter_out_entry_tags(
    entry_tag_ids: dict[EntryId, set[int]], must_have_tags: set[int], min_tag_count: int
) -> set[int]:
    return _inplace_filter_out_entry_tags(  # type: ignore
        entry_tag_ids=entry_tag_ids,  # type: ignore
        must_have_tags=must_have_tags,  # type: ignore
        min_tag_count=min_tag_count,
    )


class TestInplaceFilterOutEntryTags:

    @pytest.mark.parametrize("min_tag_count", [0, 1, 10])
    @pytest.mark.parametrize("must_have_tags", [set(), {1}, {1, 2, 10}])
    @pytest.mark.asyncio
    async def test_no_entries(self, min_tag_count: int, must_have_tags: set[int]) -> None:
        entry_tag_ids: dict[EntryId, set[int]] = {}

        whole_tags = _no_type_inplace_filter_out_entry_tags(
            entry_tag_ids=entry_tag_ids, must_have_tags=must_have_tags, min_tag_count=min_tag_count
        )

        assert whole_tags == set()
        assert entry_tag_ids == {}

    @pytest.mark.parametrize("min_tag_count", [0, 1, 10])
    @pytest.mark.parametrize("must_have_tags", [set(), {1}, {1, 2, 10}])
    @pytest.mark.asyncio
    async def test_no_tags(self, min_tag_count: int, must_have_tags: set[int]) -> None:
        entry_ids = [new_entry_id(), new_entry_id()]

        entry_tag_ids: dict[EntryId, set[int]] = {
            entry_ids[0]: set(),
            entry_ids[1]: set(),
        }

        whole_tags = _no_type_inplace_filter_out_entry_tags(
            entry_tag_ids=entry_tag_ids, must_have_tags=must_have_tags, min_tag_count=min_tag_count
        )

        assert whole_tags == set()
        assert entry_tag_ids == {
            entry_ids[0]: set(),
            entry_ids[1]: set(),
        }

    @pytest.mark.parametrize("must_have_tags", [set(), {100}, {100, 200}])
    @pytest.mark.asyncio
    async def test_filter_out_tags(self, must_have_tags: set[int]) -> None:
        entry_ids = [new_entry_id() for _ in range(5)]
        original_entry_tag_ids = {
            entry_ids[0]: {1, 2, 3, 4},
            entry_ids[1]: {2, 3, 4},
            entry_ids[2]: {3, 4, 5},
            entry_ids[3]: set(),
            entry_ids[4]: {5, 6},
        }

        filtering_no = copy.deepcopy(original_entry_tag_ids)
        filtering_1 = copy.deepcopy(original_entry_tag_ids)
        filtering_2 = copy.deepcopy(original_entry_tag_ids)
        filtering_3 = copy.deepcopy(original_entry_tag_ids)
        filtering_10 = copy.deepcopy(original_entry_tag_ids)

        whole_tags = _no_type_inplace_filter_out_entry_tags(
            entry_tag_ids=filtering_no, must_have_tags=must_have_tags, min_tag_count=0
        )
        assert whole_tags == {1, 2, 3, 4, 5, 6}
        assert filtering_no == {
            entry_ids[0]: {1, 2, 3, 4},
            entry_ids[1]: {2, 3, 4},
            entry_ids[2]: {3, 4, 5},
            entry_ids[3]: set(),
            entry_ids[4]: {5, 6},
        }

        whole_tags = _no_type_inplace_filter_out_entry_tags(
            entry_tag_ids=filtering_1, must_have_tags=must_have_tags, min_tag_count=1
        )
        assert whole_tags == {1, 2, 3, 4, 5, 6}
        assert filtering_no == {
            entry_ids[0]: {1, 2, 3, 4},
            entry_ids[1]: {2, 3, 4},
            entry_ids[2]: {3, 4, 5},
            entry_ids[3]: set(),
            entry_ids[4]: {5, 6},
        }

        whole_tags = _no_type_inplace_filter_out_entry_tags(
            entry_tag_ids=filtering_2, must_have_tags=must_have_tags, min_tag_count=2
        )
        assert whole_tags == {2, 3, 4, 5}
        assert filtering_2 == {
            entry_ids[0]: {2, 3, 4},
            entry_ids[1]: {2, 3, 4},
            entry_ids[2]: {3, 4, 5},
            entry_ids[3]: set(),
            entry_ids[4]: {5},
        }

        whole_tags = _no_type_inplace_filter_out_entry_tags(
            entry_tag_ids=filtering_3, must_have_tags=must_have_tags, min_tag_count=3
        )
        assert whole_tags == {3, 4}
        assert filtering_3 == {
            entry_ids[0]: {3, 4},
            entry_ids[1]: {3, 4},
            entry_ids[2]: {3, 4},
            entry_ids[3]: set(),
            entry_ids[4]: set(),
        }

        whole_tags = _no_type_inplace_filter_out_entry_tags(
            entry_tag_ids=filtering_10, must_have_tags=must_have_tags, min_tag_count=10
        )
        assert whole_tags == set()
        assert filtering_10 == {
            entry_ids[0]: set(),
            entry_ids[1]: set(),
            entry_ids[2]: set(),
            entry_ids[3]: set(),
            entry_ids[4]: set(),
        }

    @pytest.mark.asyncio
    async def test_keep_must_have_tags(self) -> None:
        must_have_tags = {1, 4}

        entry_ids = [new_entry_id() for _ in range(5)]
        original_entry_tag_ids = {
            entry_ids[0]: {1, 2, 3, 4},
            entry_ids[1]: {2, 3, 4},
            entry_ids[2]: {3, 4, 5},
            entry_ids[3]: set(),
            entry_ids[4]: {5, 6},
        }

        filtering_no = copy.deepcopy(original_entry_tag_ids)
        filtering_1 = copy.deepcopy(original_entry_tag_ids)
        filtering_2 = copy.deepcopy(original_entry_tag_ids)
        filtering_3 = copy.deepcopy(original_entry_tag_ids)
        filtering_10 = copy.deepcopy(original_entry_tag_ids)

        whole_tags = _no_type_inplace_filter_out_entry_tags(
            entry_tag_ids=filtering_no, must_have_tags=must_have_tags, min_tag_count=0
        )
        assert whole_tags == {1, 2, 3, 4, 5, 6}
        assert filtering_no == {
            entry_ids[0]: {1, 2, 3, 4},
            entry_ids[1]: {2, 3, 4},
            entry_ids[2]: {3, 4, 5},
            entry_ids[3]: set(),
            entry_ids[4]: {5, 6},
        }

        whole_tags = _no_type_inplace_filter_out_entry_tags(
            entry_tag_ids=filtering_1, must_have_tags=must_have_tags, min_tag_count=1
        )
        assert whole_tags == {1, 2, 3, 4, 5, 6}
        assert filtering_no == {
            entry_ids[0]: {1, 2, 3, 4},
            entry_ids[1]: {2, 3, 4},
            entry_ids[2]: {3, 4, 5},
            entry_ids[3]: set(),
            entry_ids[4]: {5, 6},
        }

        whole_tags = _no_type_inplace_filter_out_entry_tags(
            entry_tag_ids=filtering_2, must_have_tags=must_have_tags, min_tag_count=2
        )
        assert whole_tags == {1, 2, 3, 4, 5}
        assert filtering_2 == {
            entry_ids[0]: {1, 2, 3, 4},
            entry_ids[1]: {2, 3, 4},
            entry_ids[2]: {3, 4, 5},
            entry_ids[3]: set(),
            entry_ids[4]: {5},
        }

        whole_tags = _no_type_inplace_filter_out_entry_tags(
            entry_tag_ids=filtering_3, must_have_tags=must_have_tags, min_tag_count=3
        )
        assert whole_tags == {1, 3, 4}
        assert filtering_3 == {
            entry_ids[0]: {1, 3, 4},
            entry_ids[1]: {3, 4},
            entry_ids[2]: {3, 4},
            entry_ids[3]: set(),
            entry_ids[4]: set(),
        }

        whole_tags = _no_type_inplace_filter_out_entry_tags(
            entry_tag_ids=filtering_10, must_have_tags=must_have_tags, min_tag_count=10
        )
        assert whole_tags == {1, 4}
        assert filtering_10 == {
            entry_ids[0]: {1, 4},
            entry_ids[1]: {4},
            entry_ids[2]: {4},
            entry_ids[3]: set(),
            entry_ids[4]: set(),
        }


class TestPrepareTagsForEntries:

    @pytest.mark.asyncio
    async def test_no_entries(self) -> None:
        entry_tag_ids, tag_mapping = await prepare_tags_for_entries(
            entry_ids=[], must_have_tags=set(), min_tag_count=1
        )
        assert entry_tag_ids == {}
        assert tag_mapping == {}

    @pytest.mark.asyncio
    async def test_no_tags(self, cataloged_entry: Entry, another_cataloged_entry: Entry) -> None:
        entry_tag_ids, tag_mapping = await prepare_tags_for_entries(
            entry_ids=[cataloged_entry.id, another_cataloged_entry.id], must_have_tags=set(), min_tag_count=1
        )
        assert entry_tag_ids == {}
        assert tag_mapping == {}

    @pytest.mark.asyncio
    async def test_works(
        self,
        cataloged_entry: Entry,
        another_cataloged_entry: Entry,
        fake_processor_id: int,
        five_processor_tags: tuple[ProcessorTag, ProcessorTag, ProcessorTag, ProcessorTag, ProcessorTag],
    ) -> None:
        tags = five_processor_tags

        raw_to_uids = {tag.raw_uid: converters.normalize(tag.raw_uid) for tag in tags}
        uids_to_ids = await get_ids_by_uids(raw_to_uids.values())

        tag_ids = [uids_to_ids[raw_to_uids[tag.raw_uid]] for tag in tags]

        await apply_tags_to_entry(cataloged_entry.id, fake_processor_id, [tags[0], tags[1], tags[3], tags[4]])
        await apply_tags_to_entry(another_cataloged_entry.id, fake_processor_id, [tags[0], tags[2], tags[4]])

        entry_tag_ids, tag_mapping = await prepare_tags_for_entries(
            entry_ids=[cataloged_entry.id, another_cataloged_entry.id], must_have_tags=set(), min_tag_count=1
        )
        assert entry_tag_ids == {
            cataloged_entry.id: {tag_ids[0], tag_ids[1], tag_ids[3], tag_ids[4]},
            another_cataloged_entry.id: {tag_ids[0], tag_ids[2], tag_ids[4]},
        }
        assert tag_mapping == {tag_id: tag_uid for tag_uid, tag_id in uids_to_ids.items()}

    @pytest.mark.asyncio
    async def test_filtering_works(
        self,
        cataloged_entry: Entry,
        another_cataloged_entry: Entry,
        fake_processor_id: int,
        five_processor_tags: tuple[ProcessorTag, ProcessorTag, ProcessorTag, ProcessorTag, ProcessorTag],
    ) -> None:
        tags = five_processor_tags

        raw_to_uids = {tag.raw_uid: converters.normalize(tag.raw_uid) for tag in tags}
        uids_to_ids = await get_ids_by_uids(raw_to_uids.values())

        tag_ids = [uids_to_ids[raw_to_uids[tag.raw_uid]] for tag in tags]

        await apply_tags_to_entry(cataloged_entry.id, fake_processor_id, [tags[0], tags[1], tags[3], tags[4]])
        await apply_tags_to_entry(another_cataloged_entry.id, fake_processor_id, [tags[0], tags[2], tags[4]])

        entry_tag_ids, tag_mapping = await prepare_tags_for_entries(
            entry_ids=[cataloged_entry.id, another_cataloged_entry.id], must_have_tags={tag_ids[3]}, min_tag_count=2
        )
        assert entry_tag_ids == {
            cataloged_entry.id: {tag_ids[0], tag_ids[3], tag_ids[4]},
            another_cataloged_entry.id: {tag_ids[0], tag_ids[4]},
        }
        assert tag_mapping == {
            tag_ids[0]: raw_to_uids[tags[0].raw_uid],
            tag_ids[3]: raw_to_uids[tags[3].raw_uid],
            tag_ids[4]: raw_to_uids[tags[4].raw_uid],
        }



import asyncio
import uuid
from typing import Any

import pytest

from ffun.core import errors
from ffun.core.logging import (
    ArgumentConstructor,
    IdentityConstructor,
    bound_log_args,
    bound_measure_labels,
    function_args_to_log,
    get_module_logger,
)
from ffun.core.tests.helpers import assert_log_context_vars, assert_logs_has_business_event, capture_logs
from ffun.domain.domain import new_user_id

logger = get_module_logger()


class X:
    def __init__(self, my_arg: str) -> None:
        self.my_arg = my_arg


class TestMeasuringBoundLoggerMixin:
    """Test mixin methods after they applied to the logger class.

    Because it is more convinient, rather than testing the mixin itself.
    """

    def test_measure__no_labels(self) -> None:
        with capture_logs() as logs:
            logger.measure("my_event", 42)
            assert_log_context_vars()

        assert logs == [
            {
                "module": "ffun.core.tests.test_logging",
                "m_kind": "measure",
                "m_value": 42,
                "event": "my_event",
                "log_level": "info",
            }
        ]

    def test_measure__has_labels(self) -> None:
        with capture_logs() as logs:
            logger.measure("my_event", 42, x="a", y=2)

        assert logs == [
            {
                "module": "ffun.core.tests.test_logging",
                "m_kind": "measure",
                "m_value": 42,
                "event": "my_event",
                "log_level": "info",
                "m_labels": {"x": "a", "y": 2},
            }
        ]

    @pytest.mark.asyncio
    async def test_measure_block_time__no_labels(self) -> None:
        delta = 0.1

        with capture_logs() as logs:
            with logger.measure_block_time("my_event"):
                await asyncio.sleep(delta)

        assert delta == pytest.approx(logs[0]["m_value"], 0.05)
        assert logs == [
            {
                "module": "ffun.core.tests.test_logging",
                "m_kind": "measure",
                "m_value": logs[0]["m_value"],
                "event": "my_event",
                "log_level": "info",
            }
        ]

    @pytest.mark.asyncio
    async def test_measure_block_time__has_labels(self) -> None:
        delta = 0.1

        with capture_logs() as logs:
            with logger.measure_block_time("my_event", x="a", y=2):
                await asyncio.sleep(delta)
                assert_log_context_vars(m_labels={"x": "a", "y": 2})

        assert delta == pytest.approx(logs[0]["m_value"], 0.05)
        assert logs == [
            {
                "module": "ffun.core.tests.test_logging",
                "m_kind": "measure",
                "m_value": logs[0]["m_value"],
                "event": "my_event",
                "log_level": "info",
                "m_labels": {
                    "x": "a",
                    "y": 2,
                },
            }
        ]

    @pytest.mark.asyncio
    async def test_measure_block_time__extra_labels(self) -> None:
        delta = 0.1

        with capture_logs() as logs:
            with logger.measure_block_time("my_event", x="a", y=2) as extra_labels:
                await asyncio.sleep(delta)
                extra_labels["z"] = 3

        assert delta == pytest.approx(logs[0]["m_value"], 0.05)
        assert logs == [
            {
                "module": "ffun.core.tests.test_logging",
                "m_kind": "measure",
                "m_value": logs[0]["m_value"],
                "event": "my_event",
                "log_level": "info",
                "m_labels": {"x": "a", "y": 2, "z": 3},
            }
        ]


class TestBusinessBoundLoggerMixin:
    """Test mixin methods after they applied to the logger class.

    Because it is more convinient, rather than testing the mixin itself.
    """

    def test_business_event(self) -> None:
        user_id = new_user_id()

        with capture_logs() as logs:
            logger.business_event("my_event", user_id=user_id, a="b")

        assert logs == [
            {
                "module": "ffun.core.tests.test_logging",
                "event": "my_event",
                "log_level": "info",
                "b_user_id": str(user_id),
                "b_kind": "event",
                "b_uid": logs[0]["b_uid"],
                "b_attributes": {"a": "b"},
            }
        ]

        assert uuid.UUID(logs[0]["b_uid"])

        assert_logs_has_business_event(logs, "my_event", user_id=user_id, a="b")
        assert_logs_has_business_event(logs, "my_event", user_id=user_id)

    @pytest.mark.xfail
    def test_business_event__helper_expected_to_fail_because_of_arguments(self) -> None:
        user_id = new_user_id()

        with capture_logs() as logs:
            logger.business_event("my_event", user_id=user_id, a="b")

        assert_logs_has_business_event(logs, "my_event", user_id=user_id, a="c")

    @pytest.mark.xfail
    def test_business_event__helper_expected_to_fail_because_of_user_id(self) -> None:
        with capture_logs() as logs:
            logger.business_event("my_event", user_id=new_user_id(), a="b")

        assert_logs_has_business_event(logs, "my_event", user_id=new_user_id())

    @pytest.mark.xfail
    def test_business_event__helper_expected_to_fail_because_event_name(self) -> None:
        user_id = new_user_id()

        with capture_logs() as logs:
            logger.business_event("my_event", user_id=user_id, a="b")

        assert_logs_has_business_event(logs, "wrong_event", user_id=user_id, a="b")

    @pytest.mark.parametrize(
        "in_attrs, expected",
        [
            ({}, {}),
            ({"a": "b", "c": 1, 13: 2.5, "e": None}, {"a": "b", "c": 1, "13": 2.5, "e": None}),
            ({"a": uuid.UUID("12345678-1234-5678-1234-567812345678")}, {"a": "12345678-1234-5678-1234-567812345678"}),
            (
                {"a": {"b": {"c": 1}, "d": [2, {"e": 3}]}, "f": [4, None, 6]},
                {"a": {"b": {"c": 1}, "d": [2, {"e": 3}]}, "f": [4, None, 6]},
            ),
        ],
    )
    def test_normalize_value(self, in_attrs: dict[str, Any], expected: dict[str, Any]) -> None:
        assert logger._normalize_value(in_attrs) == expected


class TestIdentityConstructor:

    def test_init(self) -> None:
        constructor = IdentityConstructor("name")

        assert constructor.name == "name"

    def test_works(self) -> None:
        constructor = IdentityConstructor("name")

        assert constructor({"name": "value"}) == "value"

    def test_no_argument(self) -> None:
        constructor = IdentityConstructor("name")

        assert constructor({"bad_name": "value"}) is None


class TestArgumentConstructor:

    def test_init(self) -> None:
        constructor = ArgumentConstructor("x.my_arg")

        assert constructor.name == "x_my_arg"
        assert constructor.key == "x"
        assert constructor.attribute == "my_arg"

    def test_works(self) -> None:
        constructor = ArgumentConstructor("x.my_arg")

        assert constructor({"x": X("abc")}) == "abc"

    def test_no_argument(self) -> None:
        constructor = ArgumentConstructor("x.my_arg")

        assert constructor({"bad_name": "value"}) is None

    def test_no_attribute(self) -> None:
        constructor = ArgumentConstructor("x.wrong_arg")

        assert constructor({"x": X("abc")}) is None


class TestFunctionArgsToLog:

    def test_sync(self) -> None:

        @function_args_to_log("y", "x.my_arg")
        def func(y: int, x: X, z: int) -> None:
            logger.info("my_event")
            assert_log_context_vars(y=1, x_my_arg="abc")

        with capture_logs() as logs:
            func(y=1, x=X("abc"), z=2)

        assert logs == [
            {
                "module": "ffun.core.tests.test_logging",
                "event": "my_event",
                "log_level": "info",
                "x_my_arg": "abc",
                "y": 1,
            }
        ]

    @pytest.mark.asyncio
    async def test_async(self) -> None:

        @function_args_to_log("y", "x.my_arg")
        async def func(y: int, x: X, z: int) -> None:
            await asyncio.sleep(0)
            logger.info("my_event")
            assert_log_context_vars(y=1, x_my_arg="abc")

        with capture_logs() as logs:
            await func(y=1, x=X("abc"), z=2)

        assert logs == [
            {
                "module": "ffun.core.tests.test_logging",
                "event": "my_event",
                "log_level": "info",
                "x_my_arg": "abc",
                "y": 1,
            }
        ]


class TestBoundLogArgs:

    def test_no_args(self) -> None:
        with capture_logs() as logs:
            with bound_log_args():
                logger.info("my_event", a="b")
                logger.measure("my_event", 42, z=3)
                assert_log_context_vars()

        assert logs == [
            {"module": "ffun.core.tests.test_logging", "event": "my_event", "log_level": "info", "a": "b"},
            {
                "module": "ffun.core.tests.test_logging",
                "m_kind": "measure",
                "m_value": 42,
                "event": "my_event",
                "log_level": "info",
                "m_labels": {"z": 3},
            },
        ]

    def test_with_args(self) -> None:
        with capture_logs() as logs:
            with bound_log_args(x=1, y="a"):
                logger.info("my_event", a="b")
                logger.measure("my_event", 42, z=3)
                assert_log_context_vars(x=1, y="a")

        assert logs == [
            {
                "module": "ffun.core.tests.test_logging",
                "event": "my_event",
                "log_level": "info",
                "a": "b",
                "x": 1,
                "y": "a",
            },
            {
                "module": "ffun.core.tests.test_logging",
                "m_kind": "measure",
                "m_value": 42,
                "event": "my_event",
                "log_level": "info",
                "x": 1,
                "y": "a",
                "m_labels": {"z": 3},
            },
        ]

    @pytest.mark.parametrize("protected", ["m_labels", "m_value", "m_kind", "b_kind"])
    def test_reserved(self, protected: str) -> None:
        with pytest.raises(errors.ReservedLogArguments):
            with bound_log_args(**{protected: 1}):
                pass

    def test_recursive(self) -> None:
        with capture_logs() as logs:
            with bound_log_args(x=1):
                with bound_log_args(y=2):
                    logger.info("my_event", a="b")
                    logger.measure("my_event", 42, z=3)
                    assert_log_context_vars(x=1, y=2)

                logger.info("my_event_2", a="c")
                logger.measure("my_event_2", 43, z=4)

                assert_log_context_vars(x=1)

            assert_log_context_vars()

        assert logs == [
            {
                "module": "ffun.core.tests.test_logging",
                "event": "my_event",
                "log_level": "info",
                "a": "b",
                "x": 1,
                "y": 2,
            },
            {
                "module": "ffun.core.tests.test_logging",
                "m_kind": "measure",
                "m_value": 42,
                "event": "my_event",
                "log_level": "info",
                "x": 1,
                "y": 2,
                "m_labels": {"z": 3},
            },
            {"module": "ffun.core.tests.test_logging", "event": "my_event_2", "log_level": "info", "a": "c", "x": 1},
            {
                "module": "ffun.core.tests.test_logging",
                "m_kind": "measure",
                "m_value": 43,
                "event": "my_event_2",
                "log_level": "info",
                "x": 1,
                "m_labels": {"z": 4},
            },
        ]

    def test_duplicated_args(self) -> None:
        with pytest.raises(errors.DuplicatedLogArguments):
            with bound_log_args(x=1):
                with bound_log_args(x=2):
                    pass


class TestBoundMeasureLabels:

    def test_no_labels(self) -> None:
        with capture_logs() as logs:
            with bound_measure_labels():
                logger.info("my_event", a="b")
                logger.measure("my_event", 42, z=3)
                assert_log_context_vars()

        assert logs == [
            {"module": "ffun.core.tests.test_logging", "event": "my_event", "log_level": "info", "a": "b"},
            {
                "module": "ffun.core.tests.test_logging",
                "m_kind": "measure",
                "m_value": 42,
                "event": "my_event",
                "log_level": "info",
                "m_labels": {"z": 3},
            },
        ]

    def test_with_labels(self) -> None:
        with capture_logs() as logs:
            with bound_measure_labels(x=1, y="a"):
                logger.info("my_event", a="b")
                logger.measure("my_event", 42, z=3)
                assert_log_context_vars(m_labels={"x": 1, "y": "a"})

        assert logs == [
            {
                "module": "ffun.core.tests.test_logging",
                "event": "my_event",
                "log_level": "info",
                "a": "b",
                "m_labels": {"x": 1, "y": "a"},
            },
            {
                "module": "ffun.core.tests.test_logging",
                "m_kind": "measure",
                "m_value": 42,
                "event": "my_event",
                "log_level": "info",
                "m_labels": {"x": 1, "y": "a", "z": 3},
            },
        ]

    def test_recursive(self) -> None:
        with capture_logs() as logs:
            with bound_measure_labels(x=1):
                assert_log_context_vars(m_labels={"x": 1})

                with bound_measure_labels(y=2):
                    logger.info("my_event", a="b")
                    logger.measure("my_event", 42, z=3)
                    assert_log_context_vars(m_labels={"x": 1, "y": 2})

                logger.info("my_event_2", a="c")
                logger.measure("my_event_2", 43, z=4)

                assert_log_context_vars(m_labels={"x": 1})

            assert_log_context_vars()

        assert logs == [
            {
                "module": "ffun.core.tests.test_logging",
                "event": "my_event",
                "log_level": "info",
                "a": "b",
                "m_labels": {"x": 1, "y": 2},
            },
            {
                "module": "ffun.core.tests.test_logging",
                "m_kind": "measure",
                "m_value": 42,
                "event": "my_event",
                "log_level": "info",
                "m_labels": {"x": 1, "y": 2, "z": 3},
            },
            {
                "module": "ffun.core.tests.test_logging",
                "event": "my_event_2",
                "log_level": "info",
                "a": "c",
                "m_labels": {"x": 1},
            },
            {
                "module": "ffun.core.tests.test_logging",
                "m_kind": "measure",
                "m_value": 43,
                "event": "my_event_2",
                "log_level": "info",
                "m_labels": {"x": 1, "z": 4},
            },
        ]

    def test_duplicated_labels(self) -> None:
        with pytest.raises(errors.DuplicatedMeasureLabels):
            with bound_measure_labels(x=1):
                with bound_measure_labels(x=2):
                    pass



import asyncio

import httpx

from ffun.auth.settings import settings as auth_settings
from ffun.cli.application import app
from ffun.core import logging

logger = logging.get_module_logger()


async def run_supertokens_create_admin(email: str, password: str) -> None:
    url = f"{auth_settings.supertokens.connection_uri}/recipe/dashboard/user"

    async with httpx.AsyncClient() as client:
        response = await client.post(
            url,
            follow_redirects=True,
            headers={
                "rid": "dashboard",
                "api-key": auth_settings.supertokens.api_key,
                "Content-Type": "application/json",
            },
            json={"email": email, "password": password},
        )
        logger.info("status_code", status_code=response.status_code)
        logger.info("response", response.json())


@app.command()
def supertokens_create_admin(email: str, password: str) -> None:
    asyncio.run(run_supertokens_create_admin(email, password))



import pytest

from ffun.application.settings import Settings


class TestSettings:
    def test_origin_must_be_redefined_in_prod(self) -> None:
        with pytest.raises(ValueError):
            Settings(environment="prod")

        settings = Settings(environment="prod", origins=("http://localhost:3000", "http://localhost:3001"))

        assert settings.origins == ("http://localhost:3000", "http://localhost:3001")



import datetime

from ffun.core.entities import BaseEntity
from ffun.domain.entities import RuleId, TagId, UserId


class Rule(BaseEntity):
    id: RuleId
    user_id: UserId

    required_tags: set[TagId]
    excluded_tags: set[TagId]
    score: int

    created_at: datetime.datetime
    updated_at: datetime.datetime



import datetime
from itertools import chain
from unittest import mock

import pytest
import pytest_asyncio
from structlog.testing import capture_logs

from ffun.core import utils
from ffun.core.postgresql import execute
from ffun.core.tests.helpers import Delta, TableSizeDelta, TableSizeNotChanged, assert_logs, assert_times_is_near
from ffun.domain.domain import new_entry_id
from ffun.domain.entities import FeedId
from ffun.feeds import domain as f_domain
from ffun.feeds.entities import Feed
from ffun.feeds.tests import make as f_make
from ffun.library.domain import get_entry
from ffun.library.entities import Entry
from ffun.library.operations import (
    _catalog_entry,
    all_entries_iterator,
    catalog_entries,
    count_total_entries,
    find_stored_entries_for_feed,
    get_entries_after_pointer,
    get_entries_by_filter,
    get_entries_by_ids,
    get_feed_links_for_entries,
    get_orphaned_entries,
    remove_entries_by_ids,
    try_mark_as_orphanes,
    unlink_feed_tail,
    update_external_url,
)
from ffun.library.tests import helpers, make


class TestCatalogEntry:

    @pytest.mark.asyncio
    async def test_new_entry_new_feed(self, loaded_feed_id: FeedId, new_entry: Entry) -> None:
        async with TableSizeDelta("l_entries", delta=1):
            async with TableSizeDelta("l_feeds_to_entries", delta=1):
                await _catalog_entry(loaded_feed_id, new_entry)

        loaded_entry = await get_entry(new_entry.id)

        assert_times_is_near(loaded_entry.cataloged_at, utils.now())

        assert loaded_entry == new_entry.replace(cataloged_at=loaded_entry.cataloged_at)

        links = await get_feed_links_for_entries([new_entry.id])

        assert len(links[new_entry.id]) == 1

        link = links[new_entry.id][0]

        assert link.feed_id == loaded_feed_id
        assert link.entry_id == new_entry.id
        assert_times_is_near(link.created_at, utils.now())

    @pytest.mark.asyncio
    async def test_same_entry_new_feed(
        self, loaded_feed_id: FeedId, another_loaded_feed_id: FeedId, new_entry: Entry
    ) -> None:
        await _catalog_entry(loaded_feed_id, new_entry)

        async with TableSizeNotChanged("l_entries"):
            async with TableSizeDelta("l_feeds_to_entries", delta=1):
                await _catalog_entry(another_loaded_feed_id, new_entry)

        loaded_entry = await get_entry(new_entry.id)

        assert_times_is_near(loaded_entry.cataloged_at, utils.now())

        assert loaded_entry == new_entry.replace(cataloged_at=loaded_entry.cataloged_at)

        links = await get_feed_links_for_entries([new_entry.id])

        assert len(links[new_entry.id]) == 2

        link_1 = links[new_entry.id][0]

        assert link_1.feed_id == loaded_feed_id
        assert link_1.entry_id == new_entry.id
        assert_times_is_near(link_1.created_at, utils.now())

        link_2 = links[new_entry.id][1]

        assert link_2.feed_id == another_loaded_feed_id
        assert link_2.entry_id == new_entry.id
        assert_times_is_near(link_2.created_at, utils.now())

        assert link_1.created_at < link_2.created_at

    @pytest.mark.asyncio
    async def test_same_entry_same_feed(self, loaded_feed_id: FeedId, new_entry: Entry) -> None:
        await _catalog_entry(loaded_feed_id, new_entry)

        async with TableSizeNotChanged("l_entries"):
            async with TableSizeNotChanged("l_feeds_to_entries"):
                await _catalog_entry(loaded_feed_id, new_entry)

        loaded_entry = await get_entry(new_entry.id)

        assert_times_is_near(loaded_entry.cataloged_at, utils.now())

        assert loaded_entry == new_entry.replace(cataloged_at=loaded_entry.cataloged_at)

        links = await get_feed_links_for_entries([new_entry.id])

        assert len(links[new_entry.id]) == 1

        link = links[new_entry.id][0]

        assert link.feed_id == loaded_feed_id
        assert link.entry_id == new_entry.id
        assert_times_is_near(link.created_at, utils.now())

    @pytest.mark.asyncio
    async def test_new_entry_same_feed(
        self, loaded_feed_id: FeedId, new_entry: Entry, another_new_entry: Entry
    ) -> None:
        await _catalog_entry(loaded_feed_id, new_entry)

        async with TableSizeDelta("l_entries", delta=1):
            async with TableSizeDelta("l_feeds_to_entries", delta=1):
                await _catalog_entry(loaded_feed_id, another_new_entry)

        loaded_another_entry = await get_entry(another_new_entry.id)

        assert_times_is_near(loaded_another_entry.cataloged_at, utils.now())

        assert loaded_another_entry == another_new_entry.replace(cataloged_at=loaded_another_entry.cataloged_at)

        links = await get_feed_links_for_entries([another_new_entry.id])

        assert len(links[another_new_entry.id]) == 1

        link = links[another_new_entry.id][0]

        assert link.feed_id == loaded_feed_id
        assert link.entry_id == another_new_entry.id
        assert_times_is_near(link.created_at, utils.now())


class TestCatalogEntries:
    @pytest.mark.asyncio
    async def test_no_entries(self, loaded_feed_id: FeedId) -> None:
        async with TableSizeNotChanged("l_entries"):
            await catalog_entries(loaded_feed_id, [])

    @pytest.mark.asyncio
    async def test_success(self, loaded_feed_id: FeedId, new_entry: Entry, another_new_entry: Entry) -> None:
        entries_data = [new_entry, another_new_entry]

        async with TableSizeDelta("l_entries", delta=2):
            await catalog_entries(loaded_feed_id, entries_data)

        loaded_entries = await get_entries_by_ids(ids=[new_entry.id, another_new_entry.id])

        loaded_new_entry = loaded_entries[new_entry.id]
        loaded_another_new_entry = loaded_entries[another_new_entry.id]

        assert len(loaded_entries) == 2

        assert loaded_new_entry is not None
        assert loaded_another_new_entry is not None

        assert_times_is_near(loaded_new_entry.cataloged_at, utils.now())
        assert_times_is_near(loaded_another_new_entry.cataloged_at, utils.now())

        assert loaded_new_entry == new_entry.replace(cataloged_at=loaded_new_entry.cataloged_at)
        assert loaded_another_new_entry == another_new_entry.replace(
            cataloged_at=loaded_another_new_entry.cataloged_at
        )


# Most of the functionality is tested in the tests for catalog_entry and other functions
class TestGetFeedLinksForEntries:

    @pytest.mark.asyncio
    async def test_no_entries(self) -> None:
        assert await get_feed_links_for_entries([]) == {}


class TestFindStoredEntriesForFeed:
    @pytest.mark.asyncio
    async def test_no_entries_stored(self, loaded_feed: Feed) -> None:
        entries = [make.fake_entry(loaded_feed.source_id) for _ in range(3)]
        external_ids = [entry.external_id for entry in entries]

        stored_entries = await find_stored_entries_for_feed(loaded_feed.id, external_ids)

        assert stored_entries == set()

    @pytest.mark.asyncio
    async def test_all_entries_stored(self, loaded_feed: Feed) -> None:
        entries = await make.n_entries(loaded_feed, n=3)
        external_ids = {entry.external_id for entry in entries.values()}

        stored_entries = await find_stored_entries_for_feed(loaded_feed.id, list(external_ids))

        assert stored_entries == external_ids

    @pytest.mark.asyncio
    async def test_some_entries_stored(self, loaded_feed: Feed) -> None:
        new_entries = [make.fake_entry(loaded_feed.source_id) for _ in range(3)]
        saved_entries = await make.n_entries(loaded_feed, n=2)
        external_ids = [entry.external_id for entry in new_entries] + [
            entry.external_id for entry in saved_entries.values()
        ]

        stored_entries = await find_stored_entries_for_feed(loaded_feed.id, external_ids)

        assert stored_entries == set(entry.external_id for entry in saved_entries.values())

    @pytest.mark.asyncio
    async def test_some_entries_stored_in_another_feed(self, loaded_feed: Feed, another_loaded_feed: Feed) -> None:
        saved_entries = await make.n_entries_list(loaded_feed, n=2)
        saved_another_entries = await make.n_entries_list(another_loaded_feed, n=2)

        common_entry = make.fake_entry(loaded_feed.source_id)
        await catalog_entries(loaded_feed.id, [common_entry])
        await catalog_entries(another_loaded_feed.id, [common_entry])

        external_ids = [entry.external_id for entry in chain(saved_entries, saved_another_entries, [common_entry])]

        stored_entries = await find_stored_entries_for_feed(loaded_feed.id, external_ids)
        assert stored_entries == set(entry.external_id for entry in saved_entries) | {common_entry.external_id}

        stored_entries = await find_stored_entries_for_feed(another_loaded_feed.id, external_ids)
        assert stored_entries == set(entry.external_id for entry in saved_another_entries) | {common_entry.external_id}


class TestGetEntriesByIds:
    @pytest.mark.asyncio
    async def test_no_entries(self) -> None:
        entries = await get_entries_by_ids(ids=[])
        assert entries == {}

    @pytest.mark.asyncio
    async def test_success(self, loaded_feed: Feed, another_loaded_feed: Feed) -> None:
        entries = await make.n_entries(loaded_feed, n=3)
        another_entries = await make.n_entries(another_loaded_feed, n=3)

        entries_list = list(entries.values())
        another_entries_list = list(another_entries.values())

        entries_to_load = [*entries_list[:2], *another_entries_list[:2]]

        loaded_entries = await get_entries_by_ids(ids=[entry.id for entry in entries_to_load])

        assert len(loaded_entries) == 4
        assert entries_to_load[0] == loaded_entries[entries_to_load[0].id]
        assert entries_to_load[1] == loaded_entries[entries_to_load[1].id]
        assert another_entries_list[0] == loaded_entries[another_entries_list[0].id]
        assert another_entries_list[1] == loaded_entries[another_entries_list[1].id]


class TestGetEntriesByFilter:
    @pytest.fixture
    def time_border(self) -> datetime.datetime:
        return utils.now() - datetime.timedelta(days=1)

    @pytest_asyncio.fixture
    async def prepared_entries(
        self, loaded_feed: Feed, another_loaded_feed: Feed, time_border: datetime.datetime
    ) -> tuple[list[Entry], list[Entry]]:
        entries = await make.n_entries_list(loaded_feed, n=3)
        another_entries = await make.n_entries_list(another_loaded_feed, n=3)

        common_entry = make.fake_entry(loaded_feed.source_id)
        await catalog_entries(loaded_feed.id, [common_entry])
        await catalog_entries(another_loaded_feed.id, [common_entry])

        await helpers.update_cataloged_time(
            entries_ids=[entries[0].id, another_entries[0].id], new_time=time_border - datetime.timedelta(seconds=10)
        )

        all_entries = await get_entries_by_ids(
            ids=[entry.id for entry in chain(entries, another_entries, [common_entry])]
        )

        all_entries_list = list(all_entries.values())
        all_entries_list.sort(key=lambda entry: entry.cataloged_at)  # type: ignore

        entries_1 = []
        entries_2 = []

        for entry in all_entries_list:
            assert entry is not None

            if entry.id in {e.id for e in entries}:
                entries_1.append(entry)
            elif entry.id in {e.id for e in another_entries}:
                entries_2.append(entry)
            elif entry.id == common_entry.id:
                entries_1.append(entry)
                entries_2.append(entry)
            else:
                raise NotImplementedError()

        return (entries_1, entries_2)

    @pytest.mark.asyncio
    async def test_all(
        self, loaded_feed_id: FeedId, another_loaded_feed_id: FeedId, prepared_entries: tuple[list[Entry], list[Entry]]
    ) -> None:
        loaded_entries = await get_entries_by_filter(feeds_ids=[loaded_feed_id, another_loaded_feed_id], limit=100)

        loaded_entries_ids = {entry.id for entry in loaded_entries}
        assert loaded_entries_ids == {entry.id for entry in chain(*prepared_entries)}

    @pytest.mark.asyncio
    async def test_limit(
        self, loaded_feed_id: FeedId, another_loaded_feed_id: FeedId, prepared_entries: tuple[list[Entry], list[Entry]]
    ) -> None:
        loaded_entries = await get_entries_by_filter(feeds_ids=[loaded_feed_id, another_loaded_feed_id], limit=5)

        loaded_entries_ids = {entry.id for entry in loaded_entries}
        assert loaded_entries_ids == {entry.id for entry in chain(prepared_entries[0][1:], prepared_entries[1][1:])}

    @pytest.mark.asyncio
    async def test_feeds_filter(
        self, loaded_feed_id: FeedId, prepared_entries: tuple[list[Entry], list[Entry]]
    ) -> None:
        loaded_entries = await get_entries_by_filter(feeds_ids=[loaded_feed_id], limit=100)
        loaded_entries_ids = {entry.id for entry in loaded_entries}
        assert loaded_entries_ids == {entry.id for entry in prepared_entries[0]}

    @pytest.mark.asyncio
    async def test_time_period(
        self, loaded_feed_id: FeedId, another_loaded_feed_id: FeedId, prepared_entries: tuple[list[Entry], list[Entry]]
    ) -> None:
        loaded_entries = await get_entries_by_filter(
            feeds_ids=[loaded_feed_id, another_loaded_feed_id], limit=100, period=datetime.timedelta(days=1)
        )

        loaded_entries_ids = {entry.id for entry in loaded_entries}
        assert loaded_entries_ids == {entry.id for entry in chain(prepared_entries[0][1:], prepared_entries[1][1:])}


class TestGetEntriesAfterPointer:
    @pytest.mark.asyncio
    async def test_no_entries(self) -> None:
        entries = await get_entries_after_pointer(created_at=utils.now(), entry_id=new_entry_id(), limit=100)
        assert entries == []

    @pytest.mark.asyncio
    async def test_get_some(self, loaded_feed: Feed) -> None:
        enries = await make.n_entries(loaded_feed, n=5)

        entries_list = list(enries.values())
        entries_list.sort(key=lambda entry: (entry.cataloged_at, entry.id))

        loaded_entries = await get_entries_after_pointer(
            created_at=entries_list[2].cataloged_at, entry_id=entries_list[2].id, limit=100
        )

        assert [(entry.id, entry.cataloged_at) for entry in entries_list[3:]] == loaded_entries

    @pytest.mark.asyncio
    async def test_duplicated_created_at(self, loaded_feed: Feed) -> None:
        entries = await make.n_entries(loaded_feed, n=5)

        entries_list = list(entries.values())

        base_time = entries_list[1].cataloged_at

        await execute(
            "UPDATE l_entries SET created_at = %(created_at)s WHERE id = ANY(%(ids)s)",
            {"created_at": base_time, "ids": [entry.id for entry in entries_list[1:5]]},
        )

        entries = await get_entries_by_ids(ids=[entry.id for entry in entries_list])  # type: ignore

        entries_list = list(entries.values())
        entries_list.sort(key=lambda entry: (entry.cataloged_at, entry.id))

        for i in range(1, 4):
            loaded_entries = await get_entries_after_pointer(
                created_at=base_time, entry_id=entries_list[i].id, limit=100
            )

            assert [(entry.id, entry.cataloged_at) for entry in entries_list[i + 1 :]] == loaded_entries

    @pytest.mark.asyncio
    async def test_limit(self, loaded_feed: Feed) -> None:
        entries = await make.n_entries(loaded_feed, n=5)

        entries_list = list(entries.values())

        entries_list.sort(key=lambda entry: (entry.cataloged_at, entry.id))

        loaded_entries = await get_entries_after_pointer(
            created_at=entries_list[2].cataloged_at, entry_id=entries_list[2].id, limit=2
        )

        assert [(entry.id, entry.cataloged_at) for entry in entries_list[3:5]] == loaded_entries


class TestAllEntriesIterator:
    @pytest.mark.parametrize("chunk", [1, 2, 3, 4, 5, 6, 7])
    @pytest.mark.asyncio
    async def test(self, chunk: int) -> None:
        feed_1_data = await f_make.fake_feed()
        feed_1_id = await f_domain.save_feed(feed_1_data)

        feed_2_data = await f_make.fake_feed()
        feed_2_id = await f_domain.save_feed(feed_2_data)

        entries_1_data = [make.fake_entry(feed_1_data.source_id) for _ in range(3)]
        entries_2_data = [make.fake_entry(feed_2_data.source_id) for _ in range(3)]

        await catalog_entries(feed_1_id, entries_1_data)
        await catalog_entries(feed_2_id, entries_2_data)

        ids = [e.id for e in chain(entries_1_data, entries_2_data)]

        ids.sort()

        found_ids = [entry.id async for entry in all_entries_iterator(chunk=chunk) if entry.id in ids]

        assert found_ids == ids


class TestUpdateExternalUrl:
    @pytest.mark.asyncio
    async def test(self, cataloged_entry: Entry, another_cataloged_entry: Entry) -> None:
        new_url = make.fake_url()

        assert cataloged_entry.external_url != new_url

        await update_external_url(cataloged_entry.id, new_url)

        loaded_entry = await get_entry(cataloged_entry.id)
        assert loaded_entry.external_url == new_url

        loaded_another_entry = await get_entry(another_cataloged_entry.id)
        assert loaded_another_entry.external_url == another_cataloged_entry.external_url


class TestUnlinkFeedTail:

    @pytest.mark.asyncio
    async def test_zero_head(self, loaded_feed: Feed) -> None:
        await make.n_entries(loaded_feed, n=5)

        with capture_logs() as logs:
            async with TableSizeDelta("l_feeds_to_entries", delta=-5):
                async with TableSizeNotChanged("l_entries"):
                    await unlink_feed_tail(loaded_feed.id, offset=0)

        assert_logs(logs, feed_has_no_entries_tail=0, feed_entries_tail_removed=1)

        feed_entries = await get_entries_by_filter(feeds_ids=[loaded_feed.id], limit=100)

        assert set() == {entry.id for entry in feed_entries}

    @pytest.mark.asyncio
    async def test_not_excceed_limit(self, loaded_feed: Feed) -> None:
        entries = await make.n_entries_list(loaded_feed, n=5)

        with capture_logs() as logs:
            async with TableSizeNotChanged("l_feeds_to_entries"):
                async with TableSizeNotChanged("l_entries"):
                    await unlink_feed_tail(loaded_feed.id, offset=10)

        assert_logs(logs, feed_has_no_entries_tail=1, feed_entries_tail_removed=0)

        feed_entries = await get_entries_by_filter(feeds_ids=[loaded_feed.id], limit=100)

        assert {entry.id for entry in feed_entries} == {entry.id for entry in entries}

    @pytest.mark.asyncio
    async def test_limit(self, loaded_feed: Feed) -> None:
        entries = await make.n_entries_list(loaded_feed, n=15)

        with capture_logs() as logs:
            async with TableSizeDelta("l_feeds_to_entries", delta=-5):
                async with TableSizeNotChanged("l_entries"):
                    await unlink_feed_tail(loaded_feed.id, offset=10)

        assert_logs(logs, feed_has_no_entries_tail=0, feed_entries_tail_removed=1)

        feed_entries = await get_entries_by_filter(feeds_ids=[loaded_feed.id], limit=100)

        assert {entry.id for entry in feed_entries} == {entry.id for entry in entries[:10]}

        orphaned_entries = await get_orphaned_entries(limit=100500)

        assert orphaned_entries & {entry.id for entry in entries} == {entry.id for entry in entries[10:]}


class TestRemoveEntriesByIds:

    @pytest.mark.asyncio
    async def test_no_entries_in_request(self) -> None:
        async with TableSizeNotChanged("l_orphaned_entries"):
            await remove_entries_by_ids([new_entry_id()])

    @pytest.mark.asyncio
    async def test_no_entries_to_remove(self) -> None:
        async with TableSizeNotChanged("l_orphaned_entries"):
            await remove_entries_by_ids([])

    @pytest.mark.asyncio
    async def test_remove(self, loaded_feed: Feed, another_loaded_feed: Feed) -> None:
        entries = await make.n_entries_list(loaded_feed, n=10)

        await catalog_entries(another_loaded_feed.id, entries[2:7])

        await unlink_feed_tail(loaded_feed.id, 6)

        # 0,1 entries goes to orphanes
        # 2,3 is linked only to another_loaded_feed
        # 4,5,6 is linked to both feeds
        # 7,8,9 is linked only to loaded_feed

        entries_to_remove = [entries[0].id, entries[3].id, entries[5].id, entries[6].id, entries[8].id]

        async with TableSizeDelta("l_orphaned_entries", delta=-1):
            async with TableSizeDelta("l_feeds_to_entries", delta=-6):
                async with TableSizeDelta("l_entries", delta=-5):
                    await remove_entries_by_ids(entries_to_remove)


class TestGetOrphanedEntries:

    @pytest_asyncio.fixture(autouse=True)
    async def cleanup_orphaned_entries(self) -> None:
        await execute("DELETE FROM l_orphaned_entries")

    @pytest.mark.asyncio
    async def test_no_orphaned_entries(self) -> None:
        assert await get_orphaned_entries(limit=100) == set()

    @pytest.mark.asyncio
    async def test_zero_limit(self) -> None:
        assert await get_orphaned_entries(limit=0) == set()

    @pytest.mark.asyncio
    async def test_orphaned_entries(self, loaded_feed: Feed) -> None:
        entries = await make.n_entries_list(loaded_feed, n=5)

        await unlink_feed_tail(loaded_feed.id, 2)

        assert await get_orphaned_entries(limit=100) == {entry.id for entry in entries[2:]}

    @pytest.mark.asyncio
    async def test_limit(self, loaded_feed: Feed) -> None:
        entries = await make.n_entries_list(loaded_feed, n=5)

        await unlink_feed_tail(loaded_feed.id, 1)

        orphaned_entries = await get_orphaned_entries(limit=2)

        assert len(orphaned_entries) == 2

        assert len(orphaned_entries & {entry.id for entry in entries[1:]}) == 2


class TestTryMarkAsOrphanes:

    @pytest.mark.asyncio
    async def test_no_entries(self) -> None:
        async with TableSizeNotChanged("l_orphaned_entries"):
            await try_mark_as_orphanes([])

    @pytest.mark.asyncio
    async def test_no_orphanes_found(self, loaded_feed: Feed) -> None:
        entries = await make.n_entries_list(loaded_feed, n=3)

        async with TableSizeNotChanged("l_orphaned_entries"):
            await try_mark_as_orphanes([entry.id for entry in entries])

    @pytest.mark.asyncio
    async def test_orphanes_found(self, loaded_feed: Feed) -> None:
        entries = await make.n_entries_list(loaded_feed, n=5)

        with mock.patch("ffun.library.operations.try_mark_as_orphanes"):
            await unlink_feed_tail(loaded_feed.id, 3)

        async with TableSizeDelta("l_orphaned_entries", delta=2):
            await try_mark_as_orphanes([entry.id for entry in entries])

        orphaned_entries = await get_orphaned_entries(limit=100500)

        assert orphaned_entries & {entry.id for entry in entries} == {entry.id for entry in entries[3:]}


class TestCountTotalEntries:

    @pytest.mark.asyncio
    async def test(self, loaded_feed: Feed) -> None:
        async with Delta(count_total_entries, delta=3):
            await make.n_entries_list(loaded_feed, n=3)



import datetime
import enum
from typing import Any

from ffun.core import utils
from ffun.core.entities import BaseEntity
from ffun.domain.entities import AbsoluteUrl, EntryId, FeedId, SourceId


class ProcessedState(int, enum.Enum):
    success = 1
    error = 2
    retry_later = 3


class Entry(BaseEntity):
    id: EntryId
    source_id: SourceId
    title: str
    body: str
    external_id: str
    external_url: AbsoluteUrl
    external_tags: set[str]
    published_at: datetime.datetime

    # TODO: we do not save this field to the database
    #       at the saving time it set to NOW()
    #       => a value that is set up on the loading time will be lost
    #       maybe, it will be a good idea to have a two separate classes:
    #       - LoadedEntry — only for loader
    #       - SavedEntry — for everyone else
    cataloged_at: datetime.datetime

    @property
    def age(self) -> datetime.timedelta:
        return utils.now() - self.published_at

    def log_info(self) -> dict[str, Any]:
        return {"id": self.id, "source_id": self.source_id, "title": self.title, "external_url": self.external_url}


class EntryChange(BaseEntity):
    id: EntryId
    field: str
    old_value: Any
    new_value: Any


class FeedEntryLink(BaseEntity):
    feed_id: FeedId
    entry_id: EntryId
    created_at: datetime.datetime



import pytest

from ffun.librarian.background_processors import EntriesProcessor, ProcessorInfo
from ffun.librarian.entities import ProcessorType
from ffun.librarian.processors.base import AlwaysConstantProcessor


@pytest.fixture
def fake_processor_id() -> int:
    return 11042


@pytest.fixture
def another_fake_processor_id() -> int:
    return 11043


@pytest.fixture
def fake_processor_info(fake_processor_id: int) -> ProcessorInfo:
    return ProcessorInfo(
        id=fake_processor_id,
        type=ProcessorType.fake,
        processor=AlwaysConstantProcessor(
            name="fake_constant_processor", tags=["fake-constant-tag-1", "fake-constant-tag-2"]
        ),
        concurrency=5,
        allowed_for_collections=True,
        allowed_for_users=True,
    )


@pytest.fixture
def fake_entries_processor(fake_processor_info: ProcessorInfo) -> EntriesProcessor:
    return EntriesProcessor(processor_info=fake_processor_info, name="fake_entries_processor", delay_between_runs=0.1)



import asyncio
from typing import Any

from ffun.core import logging

logger = logging.get_module_logger()


class InfiniteTask:
    __slots__ = ("_stop_requested", "_runner_task", "_run_requested", "_delay_between_runs", "_scheduler", "_name")

    def __init__(self, name: str, delay_between_runs: float) -> None:
        self._name = name
        self._stop_requested: bool = False
        self._run_requested = asyncio.Event()
        self._delay_between_runs = delay_between_runs
        self._runner_task: asyncio.Task[Any] | None = None
        self._scheduler: asyncio.Task[Any] | None = None

    @property
    def name(self) -> str:
        return self._name

    @property
    def running(self) -> bool:
        return self._runner_task is not None

    @property
    def stop_requested(self) -> bool:
        return self._stop_requested

    @property
    def delay_between_runs(self) -> float:
        return self._delay_between_runs

    def request_run(self) -> None:
        self._run_requested.set()

    async def single_run(self) -> None:
        raise NotImplementedError('You must implement "single_run" method in child class')

    def start(self, from_start: bool = False) -> None:
        if self._runner_task is not None:
            raise NotImplementedError("You can not start task twice")

        self._stop_requested = False
        self._runner_task = asyncio.create_task(self._run(), name=f"{self._name}_runner")

        if from_start:
            self._run_requested.set()

        self._scheduler = asyncio.create_task(self._scheduler_loop(), name=f"{self._name}_scheduler")

    async def _scheduler_loop(self) -> None:
        while not self._stop_requested:
            await asyncio.sleep(self._delay_between_runs)
            self._run_requested.set()

    def request_stop(self) -> None:
        if not self.running:
            return

        self._stop_requested = True
        self._run_requested.set()

    async def stop(self) -> None:
        self.request_stop()

        if self._runner_task:
            await self._runner_task
            self._runner_task = None

        if self._scheduler:
            await self._scheduler
            self._scheduler = None

        self._stop_requested = False

    async def _run(self) -> None:
        logger.info("start_background_task", task_name=self._name)

        while not self._stop_requested:
            try:
                await self._run_requested.wait()

                self._run_requested.clear()

                logger.debug("running_background_task", task_name=self._name)

                await self.single_run()

                if self._stop_requested:
                    logger.info("stop_background_task", task_name=self._name)
                    break

            except Exception:
                logger.exception("error_in_background_task", task_name=self._name)

                # something bad happened with the infrastructure
                # there are now points in continue working
                # but we can not wait here for stop, because stop is waiting for this method
                self.request_stop()



"""
remove-l-entry-process-info
"""

from typing import Any

from psycopg import Connection
from yoyo import step

__depends__ = {"20240315_01_tWftt-optimize-l-entries", "20240319_01_MTJyn-migrate-to-processors-queue"}


sql_create_entry_process_info_table = """
CREATE TABLE l_entry_process_info (
    entry_id UUID NOT NULL REFERENCES l_entries (id),
    processor_id INTEGER NOT NULL,
    created_at TIMESTAMP WITH TIME ZONE NOT NULL DEFAULT NOW(),
    updated_at TIMESTAMP WITH TIME ZONE NOT NULL DEFAULT NOW(),
    processed_at TIMESTAMP WITH TIME ZONE NULL,
    state SMALLINT NOT NULL DEFAULT 1,
    last_error TEXT,

    PRIMARY KEY (processor_id, entry_id)
)
"""


def apply_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute("DROP TABLE l_entry_process_info")


def rollback_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute(sql_create_entry_process_info_table)


steps = [step(apply_step, rollback_step)]



import asyncio

import typer

from ffun.application.application import with_app
from ffun.core import logging
from ffun.user_settings import domain as us_domain

logger = logging.get_module_logger()

cli_app = typer.Typer()


async def run_remove_deprecated_settings() -> None:
    async with with_app():
        await us_domain.remove_deprecated_settings()


@cli_app.command()
def remove_deprecated_settings() -> None:
    asyncio.run(run_remove_deprecated_settings())



"""
unity-duplicated-feeds
"""

from typing import Any

from psycopg import Connection
from yoyo import step

__depends__ = {"20240504_02_gEapd-fill-uids-for-feeds"}


def apply_step(conn: Connection[list[Any]]) -> None:
    pass


def rollback_step(conn: Connection[dict[str, Any]]) -> None:
    pass


steps = [step(apply_step, rollback_step)]



from importlib import metadata
from typing import Any, Iterable

import fastapi
from fastapi.openapi.docs import get_swagger_ui_html
from fastapi.openapi.utils import get_openapi
from fastapi.responses import HTMLResponse, JSONResponse, PlainTextResponse

from ffun.api import entities
from ffun.api.settings import settings
from ffun.auth import domain as a_domain
from ffun.auth.dependencies import OptionalUser, User
from ffun.core import logging
from ffun.core.api import Message, MessageType
from ffun.core.errors import APIError
from ffun.data_protection import domain as dp_domain
from ffun.domain.domain import no_user_id
from ffun.domain.entities import TagId, TagUid, UserId
from ffun.domain.urls import url_to_uid
from ffun.feeds import domain as f_domain
from ffun.feeds_collections.collections import collections
from ffun.feeds_discoverer import domain as fd_domain
from ffun.feeds_discoverer import entities as fd_entities
from ffun.feeds_links import domain as fl_domain
from ffun.library import domain as l_domain
from ffun.library import entities as l_entities
from ffun.markers import domain as m_domain
from ffun.meta import domain as meta_domain
from ffun.ontology import domain as o_domain
from ffun.parsers import domain as p_domain
from ffun.parsers import entities as p_entities
from ffun.resources import domain as r_domain
from ffun.scores import domain as s_domain
from ffun.scores import entities as s_entities
from ffun.user_settings import domain as us_domain

router = fastapi.APIRouter()

logger = logging.get_module_logger()


@router.post("/api/test/internal-error")
async def api_internal_error() -> None:
    raise Exception("test_error")


@router.post("/api/test/expected-error")
async def api_expected_error() -> None:
    raise APIError(code="expected_test_error", message="Expected test error")


@router.post("/api/test/ok")
async def api_ok() -> None:
    return None


@router.post("/api/get-feeds")
async def api_get_feeds(request: entities.GetFeedsRequest, user: User) -> entities.GetFeedsResponse:
    linked_feeds = await fl_domain.get_linked_feeds(user.id)

    feeds_to_links = {link.feed_id: link for link in linked_feeds}

    feeds = await f_domain.get_feeds(ids=list(feeds_to_links.keys()))

    external_feeds = []

    for feed in feeds:
        collection_ids = collections.collections_for_feed(feed.id)
        external_feed = entities.Feed.from_internal(feed, link=feeds_to_links[feed.id], collection_ids=collection_ids)
        external_feeds.append(external_feed)

    return entities.GetFeedsResponse(feeds=external_feeds)


async def _external_entries(  # pylint: disable=R0914
    entries: Iterable[l_entities.Entry],
    with_body: bool,
    user_id: UserId | None,
    min_tag_count: int,
    include_tags: bool,
) -> tuple[list[entities.Entry], dict[TagId, TagUid]]:

    entries_ids = [entry.id for entry in entries]

    ########################
    # load rules and markers
    ########################

    if user_id is not None:
        markers = await m_domain.get_markers(user_id=user_id, entries_ids=entries_ids)
        rules = await s_domain.get_rules(user_id)
    else:
        markers = {}
        rules = []

    ##############
    # process tags
    ##############

    if include_tags:
        must_have_tags = set()
        for rule in rules:
            must_have_tags.update(rule.required_tags)
            must_have_tags.update(rule.excluded_tags)

        entry_tag_ids, tag_mapping = await o_domain.prepare_tags_for_entries(
            entry_ids=entries_ids, must_have_tags=must_have_tags, min_tag_count=min_tag_count
        )
    else:
        entry_tag_ids = {}
        tag_mapping = {}

    ####################
    # construct response
    ####################

    external_entries = []

    for entry in entries:
        score, contributions_by_ids = s_domain.get_score_contributions(rules, entry_tag_ids.get(entry.id, set()))

        external_markers = [entities.Marker.from_internal(marker) for marker in markers.get(entry.id, ())]

        external_entry = entities.Entry.from_internal(
            entry=entry,
            tags=entry_tag_ids.get(entry.id, ()),
            markers=external_markers,
            score=score,
            score_contributions=contributions_by_ids,
            with_body=with_body,
        )

        external_entries.append(external_entry)

    external_entries.sort(key=lambda entry: entry.score, reverse=True)

    return external_entries, tag_mapping


@router.post("/api/get-last-entries")
async def api_get_last_entries(request: entities.GetLastEntriesRequest, user: User) -> entities.GetLastEntriesResponse:
    linked_feeds = await fl_domain.get_linked_feeds(user.id)

    linked_feeds_ids = [link.feed_id for link in linked_feeds]

    entries = await l_domain.get_entries_by_filter_with_fallback(
        feeds_ids=linked_feeds_ids,
        period=request.period,
        limit=settings.max_returned_entries,
        fallback_limit=settings.news_outside_period,
    )

    external_entries, tags_mapping = await _external_entries(
        entries, with_body=False, user_id=user.id, min_tag_count=request.minTagCount, include_tags=True
    )

    return entities.GetLastEntriesResponse(entries=external_entries, tagsMapping=tags_mapping)


@router.post("/api/get-last-collection-entries")
async def api_get_last_collection_entries(
    request: entities.GetLastCollectionEntriesRequest,
) -> entities.GetLastCollectionEntriesResponse:

    collection = collections.collection_by_slug(request.collectionSlug)

    feed_ids = [feed_info.feed_id for feed_info in collection.feeds if feed_info.feed_id is not None]

    entries = await l_domain.get_entries_by_filter_with_fallback(
        feeds_ids=feed_ids,
        period=request.period,
        limit=settings.max_returned_entries,
        fallback_limit=settings.news_outside_period,
    )

    external_entries, tags_mapping = await _external_entries(
        entries, with_body=False, user_id=None, min_tag_count=request.minTagCount, include_tags=True
    )

    return entities.GetLastCollectionEntriesResponse(entries=external_entries, tagsMapping=tags_mapping)


@router.post("/api/get-entries-by-ids")
async def api_get_entries_by_ids(
    request: entities.GetEntriesByIdsRequest, user: OptionalUser
) -> entities.GetEntriesByIdsResponse:
    # TODO: check if belongs to user, but do not forget about public collections

    user_id = user.id if user is not None else None

    if len(request.ids) > settings.max_entries_details_requests:
        # TODO: better error processing
        raise fastapi.HTTPException(status_code=400, detail="Too many ids")

    entries = await l_domain.get_entries_by_ids(request.ids)

    found_entries = [entry for entry in entries.values() if entry is not None]

    # We cannot know here the whole distribution of tags on the user side
    # => we set min_tag_count=0
    external_entries, tags_mapping = await _external_entries(
        found_entries, with_body=True, user_id=user_id, min_tag_count=0, include_tags=request.includeTags
    )

    return entities.GetEntriesByIdsResponse(entries=external_entries, tagsMapping=tags_mapping)


@router.post("/api/create-or-update-rule")
async def api_create_or_update_rule(
    request: entities.CreateOrUpdateRuleRequest, user: User
) -> entities.CreateOrUpdateRuleResponse:
    required_tags_ids = await o_domain.get_ids_by_uids(request.requiredTags)
    excluded_tags_ids = await o_domain.get_ids_by_uids(request.excludedTags)

    await s_domain.create_or_update_rule(
        user_id=user.id,
        score=request.score,
        required_tags=required_tags_ids.values(),
        excluded_tags=excluded_tags_ids.values(),
    )

    return entities.CreateOrUpdateRuleResponse()


@router.post("/api/delete-rule")
async def api_delete_rule(request: entities.DeleteRuleRequest, user: User) -> entities.DeleteRuleResponse:
    await s_domain.delete_rule(user_id=user.id, rule_id=request.id)

    return entities.DeleteRuleResponse()


@router.post("/api/update-rule")
async def api_update_rule(request: entities.UpdateRuleRequest, user: User) -> entities.UpdateRuleResponse:
    required_tags_ids = await o_domain.get_ids_by_uids(request.requiredTags)
    excluded_tags_ids = await o_domain.get_ids_by_uids(request.excludedTags)

    await s_domain.update_rule(
        user_id=user.id,
        rule_id=request.id,
        score=request.score,
        required_tags=required_tags_ids.values(),
        excluded_tags=excluded_tags_ids.values(),
    )

    return entities.UpdateRuleResponse()


async def _prepare_rules(rules: Iterable[s_entities.Rule]) -> list[entities.Rule]:
    all_tags = set()

    for rule in rules:
        all_tags.update(rule.required_tags)
        all_tags.update(rule.excluded_tags)

    tags_mapping = await o_domain.get_tags_by_ids(all_tags)

    external_rules = [entities.Rule.from_internal(rule=rule, tags_mapping=tags_mapping) for rule in rules]

    return external_rules


@router.post("/api/get-rules")
async def api_get_rules(request: entities.GetRulesRequest, user: User) -> entities.GetRulesResponse:
    rules = await s_domain.get_rules(user_id=user.id)

    external_rules = await _prepare_rules(rules)

    return entities.GetRulesResponse(rules=external_rules)


@router.post("/api/get-score-details")
async def api_get_score_details(
    request: entities.GetScoreDetailsRequest, user: User
) -> entities.GetScoreDetailsResponse:
    entry_id = request.entryId

    rules = await s_domain.get_rules(user.id)

    tags_ids = await o_domain.get_tags_ids_for_entries([entry_id])

    rules = s_domain.get_score_rules(rules, tags_ids.get(entry_id, set()))

    external_rules = await _prepare_rules(rules)

    return entities.GetScoreDetailsResponse(rules=external_rules)


@router.post("/api/set-marker")
async def api_set_marker(request: entities.SetMarkerRequest, user: User) -> entities.SetMarkerResponse:
    await m_domain.set_marker(user_id=user.id, entry_id=request.entryId, marker=request.marker.to_internal())

    return entities.SetMarkerResponse()


@router.post("/api/remove-marker")
async def api_remove_marker(request: entities.RemoveMarkerRequest, user: User) -> entities.RemoveMarkerResponse:
    await m_domain.remove_marker(user_id=user.id, entry_id=request.entryId, marker=request.marker.to_internal())

    return entities.RemoveMarkerResponse()


@router.post("/api/discover-feeds")
async def api_discover_feeds(request: entities.DiscoverFeedsRequest, user: User) -> entities.DiscoverFeedsResponse:
    result = await fd_domain.discover(url=request.url, depth=1)

    messages = []

    # descriptive user friendly error messages
    results_to_messages = {
        fd_entities.Status.incorrect_url: "The URL has incorrect format",
        fd_entities.Status.cannot_access_url: "Cannot access the URL",
        fd_entities.Status.not_html: "Can not parse content of the page",
        fd_entities.Status.no_feeds_found: "No feeds found at the specified URL",
    }

    if result.status == fd_entities.Status.feeds_found:
        pass
    elif result.status in results_to_messages:
        messages.append(
            Message(
                type=MessageType.error,
                code=f"discover_feeds_error:{result.status.name}",
                message=results_to_messages[result.status],
            )
        )
    else:
        raise NotImplementedError(f"Unknown status: {result.status}")

    linked_feeds = await fl_domain.get_linked_feeds(user.id)
    linked_ids = {link.feed_id for link in linked_feeds}

    found_ids = await f_domain.get_feed_ids_by_uids([feed.uid for feed in result.feeds])

    for feed in result.feeds[: settings.max_feeds_suggestions_for_site]:
        feed.entries = feed.entries[: settings.max_entries_suggestions_for_site]

    external_feeds = [
        entities.FeedInfo.from_internal(feed, is_linked=feed.uid in found_ids and found_ids[feed.uid] in linked_ids)
        for feed in result.feeds
    ]

    return entities.DiscoverFeedsResponse(feeds=external_feeds, messages=messages)


@router.post("/api/add-feed")
async def api_add_feed(request: entities.AddFeedRequest, user: User) -> entities.AddFeedResponse:
    discover_result = await fd_domain.discover(url=request.url, depth=0)

    if discover_result.status != fd_entities.Status.feeds_found:
        raise fastapi.HTTPException(status_code=400, detail="Not a feed")

    feed_info = discover_result.feeds[0]

    ids = await meta_domain.add_feeds([feed_info], user.id)

    feed = await f_domain.get_feed(ids[0])

    collection_ids = collections.collections_for_feed(feed.id)

    link = await fl_domain.get_link(user.id, feed.id)

    assert link is not None

    return entities.AddFeedResponse(feed=entities.Feed.from_internal(feed, link=link, collection_ids=collection_ids))


@router.post("/api/add-opml")
async def api_add_opml(request: entities.AddOpmlRequest, user: User) -> entities.AddOpmlResponse:
    feed_infos = p_domain.parse_opml(request.content)

    await meta_domain.add_feeds(feed_infos, user.id)

    logger.business_event("opml_import", user_id=user.id, feeds_count=len(feed_infos))

    return entities.AddOpmlResponse()


@router.get("/api/get-opml")
async def api_get_opml(user: User) -> PlainTextResponse:
    linked_feeds = await fl_domain.get_linked_feeds(user.id)

    linked_feeds_ids = [link.feed_id for link in linked_feeds]

    feeds = await f_domain.get_feeds(ids=linked_feeds_ids)

    content = p_domain.create_opml(feeds=feeds)

    headers = {"Content-Disposition": "attachment; filename=feeds-fun.opml"}

    return PlainTextResponse(content=content, media_type="application/xml", headers=headers)


@router.post("/api/unsubscribe")
async def api_unsubscribe(request: entities.UnsubscribeRequest, user: User) -> entities.UnsubscribeResponse:
    await fl_domain.remove_link(user_id=user.id, feed_id=request.feedId)

    return entities.UnsubscribeResponse()


@router.post("/api/get-collections")
async def api_get_feeds_collections(
    request: entities.GetFeedsCollectionsRequest,
) -> entities.GetFeedsCollectionsResponse:

    internal_collections = collections.collections()

    collections_to_return = [entities.Collection.from_internal(collection) for collection in internal_collections]

    return entities.GetFeedsCollectionsResponse(collections=collections_to_return)


@router.post("/api/get-collection-feeds")
async def api_get_collection_feeds(
    request: entities.GetCollectionFeedsRequest, user: User
) -> entities.GetCollectionFeedsResponse:

    collection = collections.collection(request.collectionId)

    feeds = [entities.CollectionFeedInfo.from_internal(feed_info) for feed_info in collection.feeds]

    return entities.GetCollectionFeedsResponse(feeds=feeds)


@router.post("/api/subscribe-to-collections")
async def api_subscribe_to_collections(
    request: entities.SubscribeToCollectionsRequest, user: User
) -> entities.SubscribeToCollectionsResponse:
    feeds = []

    for collection_id in request.collections:
        collection = collections.collection(collection_id)

        for feed_info in collection.feeds:
            feeds.append(
                p_entities.FeedInfo(
                    url=feed_info.url,
                    title=feed_info.title,
                    description=feed_info.description,
                    entries=[],
                    uid=url_to_uid(feed_info.url),
                )
            )

    await meta_domain.add_feeds(feeds, user.id)

    return entities.SubscribeToCollectionsResponse()


@router.post("/api/get-tags-info")
async def api_get_tags_info(request: entities.GetTagsInfoRequest) -> entities.GetTagsInfoResponse:
    tags_ids = await o_domain.get_ids_by_uids(request.uids)

    info = await o_domain.get_tags_info(tags_ids.values())

    tags_info = {}

    for uid in request.uids:
        tags_info[uid] = entities.TagInfo.from_internal(info[tags_ids[uid]], uid)

    return entities.GetTagsInfoResponse(tags=tags_info)


@router.post("/api/get-resource-history")
async def api_get_resource_history(
    request: entities.GetResourceHistoryRequest, user: User
) -> entities.GetResourceHistoryResponse:
    history = await r_domain.load_resource_history(user_id=user.id, kind=request.kind.to_internal())

    return entities.GetResourceHistoryResponse(
        history=[entities.ResourceHistoryRecord.from_internal(resource) for resource in history]
    )


@router.post("/api/get-info")
async def api_get_info(request: entities.GetInfoRequest, user: User) -> entities.GetInfoResponse:
    return entities.GetInfoResponse(userId=user.id)


###############
# user settings
###############


@router.post("/api/get-user-settings")
async def api_get_user_settings(
    request: entities.GetUserSettingsRequest, user: User
) -> entities.GetUserSettingsResponse:
    from ffun.application.user_settings import UserSetting

    values = await us_domain.load_settings(user_id=user.id, kinds=[int(kind) for kind in UserSetting])

    result_values = []

    for kind, value in values.items():
        if kind == UserSetting.test_api_key:
            continue

        result_values.append(entities.UserSetting.from_internal(kind, value))

    return entities.GetUserSettingsResponse(settings=result_values)


@router.post("/api/set-user-setting")
async def api_set_user_setting(request: entities.SetUserSettingRequest, user: User) -> entities.SetUserSettingResponse:
    await us_domain.save_setting(user_id=user.id, kind=request.kind.to_internal(), value=request.value)

    return entities.SetUserSettingResponse()


@router.post("/api/track-event")
async def api_track_event(request: entities.TrackEventRequest, user: OptionalUser) -> entities.TrackEventResponse:
    attributes = request.event.model_dump()
    event = attributes.pop("name")

    user_id = user.id if user is not None else no_user_id()

    logger.business_event(event, user_id=user_id, **attributes)

    return entities.TrackEventResponse()


@router.post("/api/remove-user")
async def api_remove_user(
    request: fastapi.Request, _request: entities.RemoveUserRequest, user: User, response: fastapi.Response
) -> entities.RemoveUserResponse:

    await dp_domain.remove_user(user_id=user.id)

    await a_domain.logout_user_from_all_sessions(user_id=user.id)

    # remove all cookies to force client to detect that user is logged out
    # also it ensures that consent cookies dialog is shown again
    for cookie in request.cookies:
        response.delete_cookie(cookie)

    return entities.RemoveUserResponse()


#######################
# Swagger documentation
#######################

swagger_ui_api_parameters: dict[str, Any] = {
    "defaultModelsExpandDepth": -1,
    "defaultModelExpandDepth": 10,
}


swagger_title = "Feeds Fun API"


swagger_description = """
Greetings!

Welcome to the documentation for the Feeds Fun API.

Please note that the project is currently in its early stages of development, and as such, the API may undergo \
significant changes. We appreciate your understanding and patience during this phase.

At present, our **documentation does not include information regarding authentication**. If you wish to utilize API \
from [feeds.fun](https://feeds.fun), we recommend referring to the [supertokens](https://supertokens.com/) \
documentation for guidance on this matter. We'll improve this aspect of the documentation in the future.

For additional resources, please visit the following links:

- [Feeds Fun Website](https://feeds.fun)
- [GitHub Repository](https://github.com/Tiendil/feeds.fun)

Thank you for your interest in the Feeds Fun API. We look forward to your contributions and feedback.

"""


@router.get("/api/openapi.json", include_in_schema=False)
async def openapi(request: fastapi.Request) -> JSONResponse:
    content = get_openapi(
        title="Feeds Fun API",
        version=metadata.version("ffun"),
        description=swagger_description,
        routes=request.app.routes,
    )

    return JSONResponse(content=content)


@router.get("/api/docs", include_in_schema=False)
async def docs(request: fastapi.Request) -> HTMLResponse:
    openapi_url = request.scope.get("root_path", "") + "/api/openapi.json"

    return get_swagger_ui_html(
        openapi_url=openapi_url, title=swagger_title, swagger_ui_parameters=swagger_ui_api_parameters
    )



"""
rename-tags-names-to-uid
"""

from typing import Any

from psycopg import Connection
from yoyo import step

__depends__ = {"20230404_01_ZUBsm-ontology-tables"}


def apply_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute("ALTER TABLE o_tags RENAME COLUMN name TO uid;")


def rollback_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute("ALTER TABLE o_tags RENAME COLUMN uid TO name;")


steps = [step(apply_step, rollback_step)]



from ffun.auth import domain as a_domain
from ffun.core import logging
from ffun.domain.entities import UserId
from ffun.users import domain as u_domain

logger = logging.get_module_logger()


async def remove_user(user_id: UserId) -> None:
    auth_services_ids = await u_domain.get_user_external_ids(user_id)

    if not auth_services_ids:
        return

    for service_id, external_id in auth_services_ids.items():
        await u_domain.unlink_user(service_id, user_id)
        await a_domain.remove_user_from_external_service(service_id, external_id)

    logger.business_event("user_removed", user_id=user_id)



import datetime

import pydantic_settings

from ffun.core.settings import BaseSettings


class Settings(BaseSettings):
    delay_before_removing_orphaned_feeds: datetime.timedelta = datetime.timedelta(days=1)

    model_config = pydantic_settings.SettingsConfigDict(env_prefix="FFUN_META_")


settings = Settings()



import pytest

from ffun.librarian.background_processors import processors
from ffun.librarian.entities import ProcessorType
from ffun.librarian.processors.llm_general import Processor as LLMGeneralProcessor


@pytest.fixture(scope="session", autouse=True)
def do_not_use_real_collection_api_keys_in_tests() -> None:

    for processor in processors:
        if processor.type != ProcessorType.llm_general:
            continue

        tags_processor = processor.processor

        assert isinstance(tags_processor, LLMGeneralProcessor)

        assert tags_processor.collections_api_key is None, "You should disable the real API keys processors' settings"


@pytest.fixture(scope="session", autouse=True)
def do_not_use_real_general_api_keys_in_tests() -> None:

    for processor in processors:
        if processor.type != ProcessorType.llm_general:
            continue

        tags_processor = processor.processor

        assert isinstance(tags_processor, LLMGeneralProcessor)

        assert tags_processor.general_api_key is None, "You should disable the real API keys processors' settings"



import datetime
import uuid

import pydantic

from ffun.feeds.entities import FeedId


class FeedLink(pydantic.BaseModel):
    user_id: uuid.UUID
    feed_id: FeedId
    created_at: datetime.datetime



import uuid
from typing import NewType

UserId = NewType("UserId", uuid.UUID)
EntryId = NewType("EntryId", uuid.UUID)
FeedId = NewType("FeedId", uuid.UUID)
CollectionId = NewType("CollectionId", uuid.UUID)
CollectionSlug = NewType("CollectionSlug", str)
SourceId = NewType("SourceId", uuid.UUID)
RuleId = NewType("RuleId", uuid.UUID)
TagId = NewType("TagId", int)
TagUid = NewType("TagUid", str)

# URL types for better normalization control in code
# conversion schemas:
# UnknownUrl -> AbsoluteUrl -> FeedUrl
# AbsoluteUrl + RelativeUrl -> AbsoluteUrl -> FeedUrl
UnknownUrl = NewType("UnknownUrl", str)  # URL from external source, we know nothing about it
AbsoluteUrl = NewType("AbsoluteUrl", str)  # Normalized and fixed absolute URL, always starts with scheme or //
RelativeUrl = NewType("RelativeUrl", str)  # not normalized relative URL
FeedUrl = NewType("FeedUrl", str)

UrlUid = NewType("UrlUid", str)  # uid that was built from URL
SourceUid = NewType("SourceUid", str)  # uid that was built from URL



import httpx
import pytest
from respx.router import MockRouter

from ffun.domain.entities import AbsoluteUrl, FeedUrl, UnknownUrl
from ffun.domain.urls import str_to_feed_url
from ffun.feeds_discoverer.domain import (
    _discover_adjust_url,
    _discover_check_candidate_links,
    _discover_check_parent_urls,
    _discover_create_soup,
    _discover_extract_feed_info,
    _discover_extract_feeds_for_reddit,
    _discover_extract_feeds_from_anchors,
    _discover_extract_feeds_from_links,
    _discover_load_url,
    _discover_stop_recursion,
    _discoverers,
    discover,
)
from ffun.feeds_discoverer.entities import Context, Result, Status


class TestDiscoverAdjustUrl:

    @pytest.mark.asyncio
    async def test_wrong_url(self) -> None:
        context = Context(raw_url=UnknownUrl("wrong_url"))

        new_context, result = await _discover_adjust_url(context)

        assert new_context == context
        assert result == Result(feeds=[], status=Status.incorrect_url)

    @pytest.mark.asyncio
    async def test_good_url(self) -> None:
        context = Context(raw_url=UnknownUrl("https://example.com?"))

        new_context, result = await _discover_adjust_url(context)

        assert new_context == context.replace(url=AbsoluteUrl("https://example.com"))
        assert result is None


class TestDiscoverLoadUrl:

    @pytest.mark.asyncio
    async def test_error(self, respx_mock: MockRouter) -> None:
        respx_mock.get("/test").mock(side_effect=httpx.ConnectTimeout("some message"))

        context = Context(raw_url=UnknownUrl("http://localhost/test"), url=str_to_feed_url("http://localhost/test"))

        new_context, result = await _discover_load_url(context)

        assert new_context == context
        assert result == Result(feeds=[], status=Status.cannot_access_url)

    @pytest.mark.asyncio
    async def test_success(self, respx_mock: MockRouter) -> None:
        expected_content = "test-response"

        mocked_response = httpx.Response(200, content=expected_content)

        respx_mock.get("/test").mock(return_value=mocked_response)

        context = Context(raw_url=UnknownUrl("http://localhost/test"), url=str_to_feed_url("http://localhost/test"))

        new_context, result = await _discover_load_url(context)

        assert new_context == context.replace(content=expected_content)
        assert result is None


class TestDiscoverExtractFeedInfo:

    @pytest.mark.asyncio
    async def test_not_a_feed(self) -> None:
        context = Context(
            raw_url=UnknownUrl("http://localhost/test"),
            url=str_to_feed_url("http://localhost/test"),
            content="some text content",
        )

        new_context, result = await _discover_extract_feed_info(context)

        assert new_context == context
        assert result is None

    @pytest.mark.asyncio
    async def test_a_feed(self, raw_feed_content: str) -> None:

        context = Context(
            raw_url=UnknownUrl("http://localhost/test"),
            url=str_to_feed_url("http://localhost/test"),
            content=raw_feed_content,
        )

        new_context, result = await _discover_extract_feed_info(context)

        assert result is not None

        assert new_context == context

        assert result.status == Status.feeds_found
        assert len(result.feeds) == 1
        assert result.feeds[0].url == FeedUrl("http://localhost/test")


class TestDiscoverCreateSoup:

    @pytest.mark.xfail(reason="need to find a case when BeautifulSoup raises an exception")
    @pytest.mark.asyncio
    async def test_not_html(self) -> None:
        context = Context(raw_url=UnknownUrl("http://localhost/test"), content="some text content")

        new_context, result = await _discover_create_soup(context)

        assert new_context == context
        assert result == Result(feeds=[], status=Status.not_html)

    @pytest.mark.asyncio
    async def test_html(self, raw_feed_content: str) -> None:
        context = Context(raw_url=UnknownUrl("http://localhost/test"), content="<html></html>")

        new_context, result = await _discover_create_soup(context)

        assert new_context.soup is not None
        assert result is None


class TestDiscoverExtractFeedsFromLinks:

    @pytest.mark.asyncio
    async def test_no_links(self) -> None:
        intro_context = Context(
            raw_url=UnknownUrl("http://localhost/test"),
            url=str_to_feed_url("http://localhost/test"),
            content="<html></html>",
        )

        context, result = await _discover_create_soup(intro_context)

        new_context, result = await _discover_extract_feeds_from_links(context)

        assert new_context == context
        assert result is None

    @pytest.mark.asyncio
    async def test_links(self) -> None:
        # "author", "help", "icon", "license", "pingback", "search", "stylesheet"
        html = """
        <html>
          <head>
            <link href="http://localhost/feed1">

            <!-- ignore these -->
            <link rel="author" href="http://localhost/feed2">
            <link rel="help" href="http://localhost/feed3">
            <link rel="icon" href="http://localhost/feed4">
            <link rel="license" href="http://localhost/feed5">
            <link rel="pingback" href="http://localhost/feed6">
            <link rel="search" href="http://localhost/feed7">
            <link rel="stylesheet" href="http://localhost/feed8">

            <link rel="random-rel" href="http://localhost/feed9">
            <link rel="random-rel" href="/feed10">
            <link rel="random-rel" href="feed11">
         </head>
         <body>
           <link href="http://localhost/feed12">
           <a href="http://localhost/feed13"></a>
        </body>
        </html>
        """

        intro_context = Context(
            raw_url=UnknownUrl("http://localhost/test/xxx"),
            url=str_to_feed_url("http://localhost/test/xxx"),
            content=html,
        )

        context, result = await _discover_create_soup(intro_context)

        new_context, result = await _discover_extract_feeds_from_links(context)

        expected_links = {
            AbsoluteUrl("http://localhost/feed1"),
            AbsoluteUrl("http://localhost/feed9"),
            AbsoluteUrl("http://localhost/feed10"),
            AbsoluteUrl("http://localhost/test/feed11"),
            AbsoluteUrl("http://localhost/feed12"),
        }

        assert new_context == context.replace(candidate_urls=expected_links)
        assert result is None


class TestDiscoverExtractFeedsFromAnchors:

    @pytest.mark.asyncio
    async def test_no_anchorts(self) -> None:
        intro_context = Context(
            raw_url=UnknownUrl("http://localhost/test"),
            url=str_to_feed_url("http://localhost/test"),
            content="<html></html>",
        )

        context, result = await _discover_create_soup(intro_context)

        new_context, result = await _discover_extract_feeds_from_anchors(context)

        assert new_context == context
        assert result is None

    @pytest.mark.asyncio
    async def test_has_anchors(self) -> None:

        html = """
        <html>
          <head>
            <link href="http://localhost/feed1">
         </head>
         <body>
           <a href="http://localhost/feed2"></a>
           <a href="feed3"></a>
           <a href="/feed4"></a>
           <a href="http://localhost/feed.5.xml"></a>
           <a href="http://localhost/feed.6.rss"></a>
           <a href="http://localhost/feed.7.atom"></a>
           <a href="http://localhost/feed.8.rdf"></a>
           <a href="http://localhost/feed.9.feed"></a>
           <a href="http://localhost/feed.10.php"></a>
           <a href="http://localhost/feed.11.asp"></a>
           <a href="http://localhost/feed.12.aspx"></a>
           <a href="http://localhost/feed.13.json"></a>
           <a href="http://localhost/feed.14.cgi"></a>
         </body>
        </html>
        """

        intro_context = Context(
            raw_url=UnknownUrl("http://localhost/test/xxx"),
            url=str_to_feed_url("http://localhost/test/xxx"),
            content=html,
        )

        context, result = await _discover_create_soup(intro_context)

        new_context, result = await _discover_extract_feeds_from_anchors(context)

        expected_links = {
            AbsoluteUrl("http://localhost/feed2"),
            AbsoluteUrl("http://localhost/test/feed3"),
            AbsoluteUrl("http://localhost/feed4"),
            AbsoluteUrl("http://localhost/feed.5.xml"),
            AbsoluteUrl("http://localhost/feed.6.rss"),
            AbsoluteUrl("http://localhost/feed.7.atom"),
            AbsoluteUrl("http://localhost/feed.8.rdf"),
            AbsoluteUrl("http://localhost/feed.9.feed"),
            AbsoluteUrl("http://localhost/feed.10.php"),
            AbsoluteUrl("http://localhost/feed.11.asp"),
            AbsoluteUrl("http://localhost/feed.12.aspx"),
            AbsoluteUrl("http://localhost/feed.13.json"),
            AbsoluteUrl("http://localhost/feed.14.cgi"),
        }

        assert new_context == context.replace(candidate_urls=expected_links)
        assert result is None


class TestDiscoverCheckParentUrls:

    @pytest.mark.asyncio
    async def test_parents(self) -> None:
        context = Context(
            raw_url=UnknownUrl("http://example.com"),
            url=str_to_feed_url("http://example.com"),
            discoverers=_discoverers,
        )

        new_context, result = await _discover_check_parent_urls(context)

        assert new_context == context
        assert result is None

    @pytest.mark.asyncio
    async def test_nothing_found(self, respx_mock: MockRouter) -> None:
        respx_mock.get("/test/feed").mock(side_effect=httpx.ConnectTimeout("some message"))
        respx_mock.get("/test/").mock(side_effect=httpx.ConnectTimeout("some message"))
        respx_mock.get("/test").mock(side_effect=httpx.ConnectTimeout("some message"))
        respx_mock.get("/").mock(side_effect=httpx.ConnectTimeout("some message"))

        context = Context(
            raw_url=UnknownUrl("http://localhost/test/feed"),
            url=str_to_feed_url("http://localhost/test/feed"),
            discoverers=_discoverers,
        )

        new_context, result = await _discover_check_parent_urls(context)

        assert new_context == context.replace()
        assert result is None

    @pytest.mark.asyncio
    async def test_found(self, respx_mock: MockRouter, raw_feed_content: str, another_raw_feed_content: str) -> None:
        test_html = """
        <html>
          <head>
            <link href="http://localhost/feed1">
            <link href="http://localhost/feed4">
         </head>
         <body>
            <a href="http://localhost/feed2"></a>
            <a href="http://localhost/feed3"></a>
         </body>
        </html>
        """

        respx_mock.get("/test/abc").mock(return_value=httpx.Response(200, content="wrong content"))
        respx_mock.get("/test/").mock(return_value=httpx.Response(200, content="wrong_content"))
        respx_mock.get("/test").mock(return_value=httpx.Response(200, content=test_html))
        respx_mock.get("/").mock(return_value=httpx.Response(200, content="wrong_content"))

        respx_mock.get("/feed1").mock(return_value=httpx.Response(200, content="wrrong content"))
        respx_mock.get("/feed2").mock(return_value=httpx.Response(200, content=another_raw_feed_content))
        respx_mock.get("/feed3").mock(return_value=httpx.Response(200, content=raw_feed_content))
        respx_mock.get("/feed4").mock(return_value=httpx.Response(200, content="wrong_content"))

        context = Context(
            raw_url=UnknownUrl("http://localhost/test/abc"),
            url=str_to_feed_url("http://localhost/test/abc"),
            discoverers=_discoverers,
        )

        new_context, result = await _discover_check_parent_urls(context)

        assert new_context == context

        assert result is not None

        assert result.status == Status.feeds_found
        assert len(result.feeds) == 2
        assert {feed.url for feed in result.feeds} == {
            FeedUrl("http://localhost/feed2"),
            FeedUrl("http://localhost/feed3"),
        }


class TestDiscoverCheckCandidateLinks:

    @pytest.mark.asyncio
    async def test_no_links(self) -> None:
        context = Context(
            raw_url=UnknownUrl("http://localhost/test/xxx"), candidate_urls=set(), discoverers=_discoverers
        )

        new_context, result = await _discover_check_candidate_links(context)

        assert new_context == context
        assert result is None

    @pytest.mark.asyncio
    async def test_works__no_feeds_found(self, respx_mock: MockRouter) -> None:
        respx_mock.get("/feed1").mock(side_effect=httpx.ConnectTimeout("some message"))
        respx_mock.get("/feed2").mock(side_effect=httpx.ConnectTimeout("some message"))

        context = Context(
            raw_url=UnknownUrl("http://localhost/test"),
            candidate_urls={
                AbsoluteUrl("http://localhost/feed1"),
                AbsoluteUrl("http://localhost/feed2"),
            },
            discoverers=_discoverers,
        )

        new_context, result = await _discover_check_candidate_links(context)

        assert new_context == context.replace(candidate_urls=set())
        assert result is None

    @pytest.mark.asyncio
    async def test_works__feeds_found(
        self, respx_mock: MockRouter, raw_feed_content: str, another_raw_feed_content: str
    ) -> None:
        respx_mock.get("/feed1").mock(return_value=httpx.Response(200, content=raw_feed_content))
        respx_mock.get("/feed2").mock(return_value=httpx.Response(200, content=another_raw_feed_content))

        context = Context(
            raw_url=UnknownUrl("http://localhost/test"),
            candidate_urls={
                AbsoluteUrl("http://localhost/feed1"),
                AbsoluteUrl("http://localhost/feed2"),
            },
            discoverers=_discoverers,
        )

        new_context, result = await _discover_check_candidate_links(context)

        assert new_context == context.replace(candidate_urls=set())

        assert result is not None
        assert result.status == Status.feeds_found
        assert len(result.feeds) == 2
        assert {feed.url for feed in result.feeds} == {
            FeedUrl("http://localhost/feed1"),
            FeedUrl("http://localhost/feed2"),
        }


class TestDiscoverStopRecursion:

    @pytest.mark.asyncio
    async def test_depth_zero(self) -> None:
        context = Context(raw_url=UnknownUrl("http://localhost/test"), depth=0, discoverers=_discoverers)

        new_context, result = await _discover_stop_recursion(context)

        assert new_context == context
        assert result is not None
        assert result.status == Status.no_feeds_found

    @pytest.mark.asyncio
    async def test_depth_not_zero(self) -> None:
        context = Context(raw_url=UnknownUrl("http://localhost/test"), depth=1, discoverers=_discoverers)

        new_context, result = await _discover_stop_recursion(context)

        assert new_context == context
        assert result is None


class TestDiscoverExtractFeedsForReddit:

    @pytest.mark.asyncio
    async def test_not_reddit(self) -> None:
        context = Context(
            raw_url=UnknownUrl("http://example.com/test"), url=str_to_feed_url("http://example.com/test")
        )

        new_context, result = await _discover_extract_feeds_for_reddit(context)

        assert new_context == context
        assert result is None

    @pytest.mark.asyncio
    async def test_old_reditt(self) -> None:
        context = Context(
            raw_url=UnknownUrl("https://old.reddit.com/r/feedsfun/"),
            url=str_to_feed_url("https://old.reddit.com/r/feedsfun/"),
        )

        new_context, result = await _discover_extract_feeds_for_reddit(context)

        assert new_context == context
        assert result is None

    @pytest.mark.parametrize(
        "url,expected_url",
        [
            ("https://www.reddit.com/r/feedsfun/", "https://www.reddit.com/r/feedsfun/.rss"),
            ("https://www.reddit.com/r/feedsfun/?sd=x", "https://www.reddit.com/r/feedsfun/.rss"),
            ("https://www.reddit.com/r/feedsfun", "https://www.reddit.com/r/feedsfun/.rss"),
            ("https://reddit.com/r/feedsfun/", "https://reddit.com/r/feedsfun/.rss"),
            ("https://reddit.com/r/feedsfun", "https://reddit.com/r/feedsfun/.rss"),
            (
                "https://www.reddit.com/r/feedsfun/comments/1ho4k84/version_116_released_meet_enhanced_rules_creation/",  # noqa
                "https://www.reddit.com/r/feedsfun/.rss",
            ),
        ],
    )
    @pytest.mark.asyncio
    async def test_new_reddit(self, url: str, expected_url: FeedUrl) -> None:
        context = Context(raw_url=UnknownUrl(url), url=str_to_feed_url(url))

        new_context, result = await _discover_extract_feeds_for_reddit(context)

        assert new_context == context.replace(candidate_urls={expected_url})
        assert result is None

    @pytest.mark.asyncio
    async def test_already_reddit_rss_url(self) -> None:
        context = Context(
            raw_url=UnknownUrl("https://www.reddit.com/r/feedsfun/.rss"),
            url=str_to_feed_url("https://www.reddit.com/r/feedsfun/.rss"),
        )

        new_context, result = await _discover_extract_feeds_for_reddit(context)

        assert new_context == context
        assert result is None


def test_discoverers_list_not_changed() -> None:
    assert _discoverers == [
        _discover_adjust_url,
        _discover_extract_feeds_for_reddit,
        _discover_check_candidate_links,
        _discover_load_url,
        _discover_extract_feed_info,
        _discover_stop_recursion,
        _discover_create_soup,
        _discover_extract_feeds_from_links,
        _discover_check_candidate_links,
        _discover_extract_feeds_from_anchors,
        _discover_check_candidate_links,
        _discover_check_parent_urls,
    ]


class TestDiscover:

    @pytest.mark.asyncio
    async def test_long_run__stop_on_links(
        self, respx_mock: MockRouter, raw_feed_content: str, another_raw_feed_content: str
    ) -> None:
        test_html = """
        <html>
          <head>
            <link href="http://localhost/feed1">
            <link href="http://localhost/feed3">
            <link href="http://localhost/feed4">
         </head>
         <body>
            <a href="http://localhost/feed2"></a>
            <a href="http://localhost/feed3"></a>
         </body>
        </html>
        """

        respx_mock.get("/test").mock(return_value=httpx.Response(200, content=test_html))
        respx_mock.get("/feed1").mock(return_value=httpx.Response(200, content=raw_feed_content))
        respx_mock.get("/feed2").mock(return_value=httpx.Response(200, content=another_raw_feed_content))
        respx_mock.get("/feed3").mock(return_value=httpx.Response(200, content="wrong content"))
        respx_mock.get("/feed4").mock(return_value=httpx.Response(200, content=another_raw_feed_content))
        respx_mock.get("/").mock(return_value=httpx.Response(200, content="wrong_content"))

        result = await discover(UnknownUrl("http://localhost/test"), depth=1)

        assert result is not None
        assert result.status == Status.feeds_found
        assert len(result.feeds) == 2
        assert {feed.url for feed in result.feeds} == {
            FeedUrl("http://localhost/feed1"),
            FeedUrl("http://localhost/feed4"),
        }

    @pytest.mark.asyncio
    async def test_long_run__stop_on_anchorts(
        self, respx_mock: MockRouter, raw_feed_content: str, another_raw_feed_content: str
    ) -> None:
        test_html = """
        <html>
          <head>
            <link href="http://localhost/feed1">
            <link href="http://localhost/feed4">
         </head>
         <body>
            <a href="http://localhost/feed2"></a>
            <a href="http://localhost/feed3"></a>
         </body>
        </html>
        """

        respx_mock.get("/test").mock(return_value=httpx.Response(200, content=test_html))
        respx_mock.get("/feed1").mock(return_value=httpx.Response(200, content="wrrong content"))
        respx_mock.get("/feed2").mock(return_value=httpx.Response(200, content=another_raw_feed_content))
        respx_mock.get("/feed3").mock(return_value=httpx.Response(200, content=raw_feed_content))
        respx_mock.get("/feed4").mock(return_value=httpx.Response(200, content="wrong_content"))
        respx_mock.get("/").mock(return_value=httpx.Response(200, content="wrong_content"))

        result = await discover(UnknownUrl("http://localhost/test"), depth=1)

        assert result is not None
        assert result.status == Status.feeds_found
        assert len(result.feeds) == 2
        assert {feed.url for feed in result.feeds} == {
            FeedUrl("http://localhost/feed2"),
            FeedUrl("http://localhost/feed3"),
        }

    @pytest.mark.asyncio
    async def test_nothing_found(self, respx_mock: MockRouter) -> None:
        respx_mock.get("/test").mock(return_value=httpx.Response(200, content="wrong content"))

        result = await discover(UnknownUrl("http://localhost/test"), depth=1)

        assert result == Result(feeds=[], status=Status.no_feeds_found)

    @pytest.mark.asyncio
    async def test_found_in_parent(
        self, respx_mock: MockRouter, raw_feed_content: str, another_raw_feed_content: str
    ) -> None:
        test_html = """
        <html>
          <head>
            <link href="http://localhost/feed1">
            <link href="http://localhost/feed4">
         </head>
         <body>
            <a href="http://localhost/feed2"></a>
            <a href="http://localhost/feed3"></a>
         </body>
        </html>
        """

        respx_mock.get("/test/abc").mock(return_value=httpx.Response(200, content="wrong content"))
        respx_mock.get("/test/").mock(return_value=httpx.Response(200, content="wrong_content"))
        respx_mock.get("/test").mock(return_value=httpx.Response(200, content=test_html))
        respx_mock.get("/").mock(return_value=httpx.Response(200, content="wrong_content"))

        respx_mock.get("/feed1").mock(return_value=httpx.Response(200, content="wrrong content"))
        respx_mock.get("/feed2").mock(return_value=httpx.Response(200, content=another_raw_feed_content))
        respx_mock.get("/feed3").mock(return_value=httpx.Response(200, content=raw_feed_content))
        respx_mock.get("/feed4").mock(return_value=httpx.Response(200, content="wrong_content"))

        result = await discover(UnknownUrl("http://localhost/test/abc"), depth=1)

        assert result is not None
        assert result.status == Status.feeds_found
        assert len(result.feeds) == 2
        assert {feed.url for feed in result.feeds} == {
            FeedUrl("http://localhost/feed2"),
            FeedUrl("http://localhost/feed3"),
        }



import datetime
import time
from typing import Any, Iterable

import feedparser

from ffun.core import logging, utils
from ffun.domain import urls
from ffun.domain.entities import AbsoluteUrl, FeedUrl
from ffun.parsers.entities import EntryInfo, FeedInfo

logger = logging.get_module_logger()


def _parse_tags(tags: Iterable[dict[str, Any]]) -> set[str]:
    result = set()

    for tag in tags:
        if tag.get("label") is not None:
            result.add(tag["label"])

        elif tag.get("term") is not None:
            result.add(tag["term"])

    return result


def _should_skip(entry: Any) -> bool:
    if entry.get("link") is None:
        logger.warning("feed_does_not_has_link_field")
        return True

    return False


def _extract_published_at(entry: Any) -> datetime.datetime:
    published_at = entry.get("published_parsed")

    if published_at is not None:
        try:
            parsed = datetime.datetime.fromtimestamp(time.mktime(published_at))
        except ValueError:
            logger.warning("wrong_published_at_value", value=str(published_at))
            parsed = utils.now()

        if parsed.tzinfo is None:
            parsed = parsed.replace(tzinfo=datetime.timezone.utc)

        return parsed

    return utils.now()


# not the best solution, but it works for now
# TODO: we should use more formal way to detect uniqueness of entry
# TODO: external id must be normalized
def _extract_external_id(entry: Any) -> str:
    return entry.get("link")  # type: ignore


def _extract_external_url(entry: Any, original_url: FeedUrl) -> AbsoluteUrl | None:
    url = entry.get("link")

    return urls.adjust_external_url(url, original_url)


def parse_entry(raw_entry: Any, original_url: FeedUrl) -> EntryInfo:
    # TODO: remove all tags from title
    # TODO: extract tags from <category> tag
    published_at = _extract_published_at(raw_entry)

    return EntryInfo(
        title=raw_entry.get("title", ""),
        body=raw_entry.get("description", ""),
        external_id=_extract_external_id(raw_entry),
        external_url=_extract_external_url(raw_entry, original_url),
        external_tags=_parse_tags(raw_entry.get("tags", ())),
        published_at=published_at,
    )


def parse_feed(content: str, original_url: FeedUrl) -> FeedInfo | None:  # noqa: CCR001
    try:
        channel = feedparser.parse(content)
    except Exception:
        logger.exception("error_while_parsing_feed", original_url=original_url)
        return None

    if getattr(channel, "version", "") == "" and not channel.entries:
        return None

    feed_info = FeedInfo(
        url=original_url,
        title=channel.feed.get("title", ""),
        description=channel.feed.get("description", ""),
        entries=[],
        uid=urls.url_to_uid(original_url),
    )

    for entry in channel.entries:
        if _should_skip(entry):
            continue

        try:
            parsed_entry = parse_entry(entry, original_url)
        except Exception:
            logger.exception("error_while_parsing_feed_entry", entry=entry, original_url=original_url)
            continue

        feed_info.entries.append(parsed_entry)

    return feed_info



"""
fix-entries-unique-index
"""

from typing import Any

from psycopg import Connection
from yoyo import step

__depends__ = {"20230331_01_UsHwp-entries-table"}

constraint = "l_entries_external_id_key"
index = "l_entries_feed_id_external_id_key"
unnecessary_index = "idx_l_entries_feed_id"

# we must controll unique constraint on entries by pair feed + id because
# 1. we can not guarantee that there will be no duplicates of feeds
#    (can not guarantee perfect normalization of urls, etc)
# 2. someone can damage database by infecting with wrong entries from faked feed


def apply_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute(f"ALTER TABLE l_entries DROP CONSTRAINT {constraint}")
    cursor.execute(f"CREATE INDEX {index} ON l_entries (feed_id, external_id)")
    cursor.execute(f"DROP INDEX {unnecessary_index}")


def rollback_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute("TRUNCATE TABLE l_entry_process_info, l_entries")
    cursor.execute(f"ALTER TABLE l_entries ADD CONSTRAINT {constraint} UNIQUE (external_id)")
    cursor.execute(f"DROP INDEX {index}")
    cursor.execute(f"CREATE INDEX {unnecessary_index} ON l_entries (feed_id)")


steps = [step(apply_step, rollback_step)]



import pydantic_settings

from ffun.core import logging
from ffun.core.settings import BaseSettings

logger = logging.get_module_logger()


# this is key for development!
# MUST be changed in production
_default_secret_key = "JHBC3P8q9nhDNStpfljJ7eO09XNztJ3xWKI5l6rvL-Q="  # noqa: S105


class Settings(BaseSettings):
    secret_key: str = _default_secret_key

    model_config = pydantic_settings.SettingsConfigDict(env_prefix="FFUN_USER_SETTINGS_")


settings = Settings()


if settings.secret_key == _default_secret_key:
    logger.error("user_settings_secret_key_not_changed")



"""
proxy-states-table
"""

from typing import Any

from psycopg import Connection
from yoyo import step

__depends__: set[str] = set()


sql_create_proxy_states_table = """
CREATE TABLE lr_proxy_states (
    name TEXT PRIMARY KEY,
    state INTEGER NOT NULL,
    created_at TIMESTAMP WITH TIME ZONE NOT NULL DEFAULT CURRENT_TIMESTAMP,
    updated_at TIMESTAMP WITH TIME ZONE NOT NULL DEFAULT CURRENT_TIMESTAMP
)
"""


def apply_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()

    cursor.execute(sql_create_proxy_states_table)


def rollback_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute("DROP TABLE lr_proxy_states")


steps = [step(apply_step, rollback_step)]



import asyncio
import uuid

import typer

from ffun.application.application import with_app
from ffun.core import logging, utils
from ffun.domain.entities import CollectionId, FeedId, UserId
from ffun.feeds import domain as f_domain
from ffun.feeds_collections.collections import collections
from ffun.feeds_links import domain as fl_domain
from ffun.librarian import background_processors
from ffun.librarian import domain as ln_domain
from ffun.library import domain as l_domain
from ffun.ontology import domain as o_domain
from ffun.resources import domain as r_domain
from ffun.scores import domain as s_domain
from ffun.users import domain as u_domain

logger = logging.get_module_logger()

cli_app = typer.Typer()

################
# System metrics
################


async def system_slice_tags() -> None:
    tags_total = await o_domain.count_total_tags()

    logger.business_slice("tags_total", user_id=None, total=tags_total)

    tags_per_type = await o_domain.count_total_tags_per_type()

    logger.business_slice(
        "tags_per_type", user_id=None, **{tag_type.name: count for tag_type, count in tags_per_type.items()}
    )

    tags_per_category = await o_domain.count_total_tags_per_category()

    logger.business_slice(
        "tags_per_category",
        user_id=None,
        **{tag_category.name: count for tag_category, count in tags_per_category.items()},
    )


async def system_slice_feeds() -> None:
    feeds_total = await f_domain.count_total_feeds()

    logger.business_slice("feeds_total", user_id=None, total=feeds_total)

    feeds_per_state = await f_domain.count_total_feeds_per_state()

    logger.business_slice(
        "feeds_per_state", user_id=None, **{feed_state.name: count for feed_state, count in feeds_per_state.items()}
    )

    feeds_per_last_error = await f_domain.count_total_feeds_per_last_error()

    logger.business_slice(
        "feeds_per_last_error",
        user_id=None,
        **{feed_error.name: count for feed_error, count in feeds_per_last_error.items()},
    )


async def system_slice_entries() -> None:
    entries_total = await l_domain.count_total_entries()

    logger.business_slice("entries_total", user_id=None, total=entries_total)


async def system_slice_users() -> None:
    users_total = await u_domain.count_total_users()

    logger.business_slice("users_total", user_id=None, total=users_total)


async def system_slice_collections() -> None:
    logger.business_slice("collections_total", user_id=None, total=collections.count_total_feeds())


async def system_slice_processors() -> None:
    pointers_list = await ln_domain.get_all_pointers()
    pointers = {pointer.processor_id: pointer for pointer in pointers_list}

    processed_processors = set()

    for processor in background_processors.processors:
        processed_processors.add(processor.id)

        pointer = pointers.get(processor.id)

        if pointer is None:
            logger.business_slice(
                "processor_pointer",
                user_id=None,
                processor_id=processor.id,
                is_active=True,
                pointer_created_at=int(utils.zero_timestamp().timestamp()),
            )
            continue

        logger.business_slice(
            "processor_pointer",
            user_id=None,
            processor_id=processor.id,
            is_active=True,
            pointer_created_at=int(pointer.pointer_created_at.timestamp()),
        )

    for pointer in pointers_list:
        if pointer.processor_id in processed_processors:
            continue

        logger.business_slice(
            "processor_pointer",
            user_id=None,
            processor_id=pointer.processor_id,
            is_active=False,
            pointer_created_at=int(pointer.pointer_created_at.timestamp()),
        )


async def run_system() -> None:

    async with with_app():
        await system_slice_tags()
        await system_slice_feeds()
        await system_slice_entries()
        await system_slice_users()
        await system_slice_collections()
        await system_slice_processors()


@cli_app.command()
def system() -> None:
    asyncio.run(run_system())


##############
# User metrics
##############


async def users_slice_rules() -> None:
    rules_per_user = await s_domain.count_rules_per_user()

    for user_id, count in rules_per_user.items():
        if count == 0:
            continue
        logger.business_slice("rules_per_user", user_id=user_id, total=count)


async def count_collections_for_users() -> None:  # noqa: CCR001
    collection_users: dict[UserId, dict[CollectionId, int]] = {}

    for collection in collections.collections():
        feed_ids: list[FeedId] = [feed_info.feed_id for feed_info in collection.feeds]  # type: ignore

        counts_for_collection = await fl_domain.count_subset_feeds_per_user(feed_ids)

        for user_id, count in counts_for_collection.items():
            if user_id not in collection_users:
                collection_users[user_id] = {}

            collection_users[user_id][collection.id] = count

    for user_id, counts in collection_users.items():
        if all(count == 0 for count in counts.values()):
            continue

        attributes = {f"collection_{collection_id}": count for collection_id, count in counts.items()}
        logger.business_slice("collection_feeds_per_user", user_id=user_id, **attributes)


async def count_feeds_fun_news_for_users() -> None:  # noqa: CCR001
    feeds_fun_collection_id = CollectionId(uuid.UUID("09887b50-48b0-420a-b614-772e85617cb7"))

    feeds_fun_users: dict[UserId, dict[FeedId, int]] = {}

    for feed_info in collections.collection(feeds_fun_collection_id).feeds:
        assert feed_info.feed_id

        counts_for_feeds_fun = await fl_domain.count_subset_feeds_per_user([feed_info.feed_id])

        for user_id, count in counts_for_feeds_fun.items():
            if user_id not in feeds_fun_users:
                feeds_fun_users[user_id] = {}

            feeds_fun_users[user_id][feed_info.feed_id] = count

    for user_id, counts in feeds_fun_users.items():
        if all(count == 0 for count in counts.values()):
            continue

        attributes = {f"feed_{feed_id}": count for feed_id, count in counts.items()}
        logger.business_slice("feeds_fun_feeds_per_user", user_id=user_id, **attributes)


async def users_slice_feeds_links() -> None:  # noqa: CCR001
    # total feeds per user
    feeds_per_user = await fl_domain.count_feeds_per_user()

    for user_id, count in feeds_per_user.items():
        if count == 0:
            continue

        logger.business_slice("feeds_per_user", user_id=user_id, total=count)

    await count_collections_for_users()

    await count_feeds_fun_news_for_users()


async def users_slice_resources() -> None:  # noqa: CCR001
    from ffun.application.resources import Resource

    users: dict[UserId, dict[str, int]] = {}

    resource_by_kind = {}

    for kind in Resource:
        resource_by_kind[kind] = await r_domain.count_total_resources_per_user(kind)

    for kind, resource_per_user in resource_by_kind.items():
        for user_id, count in resource_per_user.items():
            if user_id not in users:
                users[user_id] = {}

            users[user_id][kind.name] = count

    for user_id, resources in users.items():
        if all(count == 0 for count in resources.values()):
            continue
        logger.business_slice("resources_per_user", user_id=user_id, **resources)


async def run_users() -> None:
    async with with_app():
        await users_slice_rules()
        await users_slice_feeds_links()
        await users_slice_resources()


@cli_app.command()
def users() -> None:
    asyncio.run(run_users())



from typing import Any

import pydantic

from ffun.core.entities import BaseEntity
from ffun.llms_framework.provider_interface import ChatRequest, ChatResponse


class ChatMessage(BaseEntity):
    role: str | None = None
    text: str

    def to_api(self) -> dict[str, Any]:
        data = {"parts": [{"text": self.text}]}

        if self.role is not None:
            data["role"] = self.role  # type: ignore

        return data


class GoogleChatRequest(ChatRequest):
    system: ChatMessage
    messages: list[ChatMessage]


class GoogleChatResponse(ChatResponse):
    content: str
    prompt_tokens: int
    completion_tokens: int
    total_tokens: int

    def response_content(self) -> str:
        return self.content

    def input_tokens(self) -> int:
        return self.prompt_tokens

    def output_tokens(self) -> int:
        return self.completion_tokens


class GenerationConfig(BaseEntity):
    candidates_count: int = 1
    stop_sequences: tuple[str, ...] = pydantic.Field(default_factory=tuple)
    response_mime_type: str = "text/plain"
    response_schema: dict[str, Any] | None = None
    max_output_tokens: int
    temperature: float
    top_p: float
    top_k: int | None

    def to_api(self) -> dict[str, Any]:
        return {
            "candidateCount": self.candidates_count,
            "stopSequences": self.stop_sequences,
            "responseMimeType": self.response_mime_type,
            "responseSchema": self.response_schema,
            "maxOutputTokens": self.max_output_tokens,
            "temperature": self.temperature,
            "topP": self.top_p,
            "topK": self.top_k,
        }



import pydantic_settings


class BaseSettings(pydantic_settings.BaseSettings):
    model_config = pydantic_settings.SettingsConfigDict(
        env_nested_delimiter="__",
        # TODO: smart search for .env file
        # or update to pydantic-settings when it will be released
        env_file=("../.env", ".env"),
        extra="allow",
    )



import datetime
import enum
from typing import Any

from ffun.core.entities import BaseEntity
from ffun.domain.entities import FeedId, FeedUrl, SourceId


class FeedState(int, enum.Enum):
    not_loaded = 1
    loaded = 2
    damaged = 3
    orphaned = 4


class FeedError(enum.IntEnum):
    network_unknown = 1000

    # TODO: split to network and protocol errors?
    network_no_address_associated_with_hostname = 1001
    network_name_or_service_not_known = 1002
    network_certificate_verify_failed = 1003
    network_connection_timeout = 1004
    network_read_timeout = 1005
    network_illegal_request_line = 1006
    network_non_200_status_code = 1007
    network_disconection_without_response = 1008
    network_undetected_connection_error = 1009
    network_unsupported_protocol = 1010
    network_server_breaks_connection = 1011
    network_too_many_redirects = 1013
    network_ssl_connection_error = 1014
    network_all_connection_attempts_failed = 1015
    network_received_unkomplete_body = 1016
    network_decoding_error = 1017
    network_read_error = 1018
    network_temporary_failure_in_name_resolution = 1019
    network_wrong_ssl_version = 1020
    # TODO: this error does not used now
    #       we should return it when HTTPX starts returning new location in the RemoteProtocolError
    #       after redirect failure
    #       see details in the loader.operations
    network_great_china_firewall_redirect = 1021
    network_wrong_redirect = 1022

    parsing_unknown = 2000
    parsing_base_error = 2001
    parsing_format_error = 2002
    parsing_unicode_decode_error = 2003
    parsing_feed_content_not_found = 2004

    protocol_unknown = 3000
    protocol_no_entries_in_feed = 3001

    proxy_could_not_resolve_host = 4001
    proxy_connection_refused = 4002
    proxy_connection_403 = 4003
    proxy_no_route_to_host = 4004
    proxy_connection_502 = 4005
    proxy_connection_429 = 4006


class Feed(BaseEntity):
    id: FeedId
    source_id: SourceId
    url: FeedUrl
    state: FeedState = FeedState.not_loaded
    last_error: FeedError | None = None
    load_attempted_at: datetime.datetime | None = None
    loaded_at: datetime.datetime | None = None

    title: str | None
    description: str | None

    def log_info(self) -> dict[str, Any]:
        return {"id": self.id, "state": self.state, "url": self.url, "last_error": self.last_error}



"""
updated-at-field
"""

from typing import Any

from psycopg import Connection
from yoyo import step

__depends__ = {"20230417_01_0XYOQ-scores-tables"}

sql = """
ALTER TABLE s_rules
ADD COLUMN updated_at TIMESTAMP WITH TIME ZONE NOT NULL DEFAULT NOW()
"""


def apply_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute(sql)


def rollback_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()

    sql = """
    ALTER TABLE s_rules
    DROP COLUMN updated_at
    """

    cursor.execute(sql)


steps = [step(apply_step, rollback_step)]



import functools
import pathlib

import pydantic
import pydantic_settings
import toml

from ffun.core.settings import BaseSettings
from ffun.librarian.entities import ProcessorsConfig, TagProcessor

_root = pathlib.Path(__file__).parent


class Settings(BaseSettings):
    # TODO: add to documentation/README
    tag_processors_config: pathlib.Path = _root / "fixtures" / "tag_processors.toml"

    @pydantic.computed_field  # type: ignore
    @functools.cached_property
    def tag_processors(self) -> tuple[TagProcessor, ...]:
        data = toml.loads(self.tag_processors_config.read_text())

        return ProcessorsConfig(**data).tag_processors

    model_config = pydantic_settings.SettingsConfigDict(env_prefix="FFUN_LIBRARIAN_")


settings = Settings()



from yoyo import get_backend, read_migrations
from yoyo.migrations import MigrationList

from ffun import __path__ as ffun_path
from ffun.application.settings import settings as app_settings

ffun_root = ffun_path[0]


async def apply_all() -> None:
    migrations: MigrationList = read_migrations(f"{ffun_root}/*/migrations")

    with get_backend(app_settings.postgresql.dsn_yoyo, "yoyo_migrations") as backend:
        migrations_to_apply: MigrationList = backend.to_apply(migrations)
        backend.apply_migrations(migrations_to_apply)


async def rollback_all() -> None:
    migrations: MigrationList = read_migrations(f"{ffun_root}/*/migrations")

    with get_backend(app_settings.postgresql.dsn_yoyo, "yoyo_migrations") as backend:
        migrations_to_rollback: MigrationList = backend.to_rollback(migrations)
        backend.rollback_migrations(migrations_to_rollback)



import pytest
from structlog.testing import capture_logs

from ffun.core import utils
from ffun.core.postgresql import execute
from ffun.core.tests.helpers import TableSizeDelta, TableSizeNotChanged, assert_logs
from ffun.feeds.entities import Feed
from ffun.librarian import errors, operations
from ffun.librarian.domain import (
    move_failed_entries_to_processor_queue,
    plan_processor_queue,
    process_entry,
    push_entries_and_move_pointer,
)
from ffun.librarian.entities import ProcessorPointer
from ffun.librarian.processors.base import (
    AlwaysConstantProcessor,
    AlwaysErrorProcessor,
    AlwaysSkipEntryProcessor,
    AlwaysTemporaryErrorProcessor,
)
from ffun.librarian.tests import helpers, make
from ffun.library.entities import Entry
from ffun.library.tests import make as l_make
from ffun.ontology import domain as o_domain


class TestPushEntriesAndMovePointer:
    @pytest.mark.asyncio
    async def test_no_entries(self, fake_processor_id: int) -> None:
        pointer = await operations.get_or_create_pointer(fake_processor_id)

        async with TableSizeNotChanged("ln_processors_queue"):
            await push_entries_and_move_pointer(pointer, [])

        loaded_pointer = await operations.get_or_create_pointer(fake_processor_id)

        assert loaded_pointer == pointer

    @pytest.mark.asyncio
    async def test_push_entries(
        self, fake_processor_id: int, cataloged_entry: Entry, another_cataloged_entry: Entry
    ) -> None:
        original_pointer = await operations.get_or_create_pointer(fake_processor_id)

        entries = [cataloged_entry, another_cataloged_entry]
        entries.sort(key=lambda entry: (entry.cataloged_at, entry.id))

        last_entry = entries[-1]

        next_pointer = ProcessorPointer(
            processor_id=fake_processor_id, pointer_created_at=last_entry.cataloged_at, pointer_entry_id=last_entry.id
        )

        async with TableSizeDelta("ln_processors_queue", delta=2):
            await push_entries_and_move_pointer(next_pointer, [cataloged_entry.id, another_cataloged_entry.id])

        loaded_pointer = await operations.get_or_create_pointer(fake_processor_id)

        assert loaded_pointer == next_pointer
        assert loaded_pointer != original_pointer

    @pytest.mark.asyncio
    async def test_no_new_entries(self, fake_processor_id: int) -> None:
        pointer = await make.end_processor_pointer(fake_processor_id)

        next_pointer = pointer.replace(pointer_created_at=utils.now())

        async with TableSizeNotChanged("ln_processors_queue"):
            await push_entries_and_move_pointer(next_pointer, [])

        loaded_pointer = await operations.get_or_create_pointer(fake_processor_id)

        assert loaded_pointer == next_pointer

    @pytest.mark.asyncio
    async def test_entries_found_and_moved(
        self, fake_processor_id: int, cataloged_entry: Entry, another_cataloged_entry: Entry
    ) -> None:
        entries = [cataloged_entry, another_cataloged_entry]
        entries.sort(key=lambda entry: (entry.cataloged_at, entry.id))

        pointer = await make.end_processor_pointer(fake_processor_id)

        async with TableSizeDelta("ln_processors_queue", delta=2):
            await push_entries_and_move_pointer(pointer, [cataloged_entry.id, another_cataloged_entry.id])

        loaded_pointer = await operations.get_or_create_pointer(fake_processor_id)

        assert loaded_pointer == pointer


class TestPlanProcessorQueue:
    @pytest.mark.asyncio
    async def test_no_new_entries(self, fake_processor_id: int, cataloged_entry: Entry) -> None:
        pointer = await make.end_processor_pointer(fake_processor_id)

        async with TableSizeNotChanged("ln_processors_queue"):
            await plan_processor_queue(fake_processor_id, fill_when_below=100500, chunk=100)

        loaded_pointer = await operations.get_or_create_pointer(fake_processor_id)

        assert loaded_pointer == pointer

    @pytest.mark.asyncio
    async def test_move_pointer_to_the_end(self, loaded_feed: Feed, fake_processor_id: int) -> None:
        await make.end_processor_pointer(1)

        entries = await l_make.n_entries(loaded_feed, 3)
        entries_list = list(entries.values())
        entries_list.sort(key=lambda entry: (entry.cataloged_at, entry.id))

        async with TableSizeDelta("ln_processors_queue", delta=3):
            await plan_processor_queue(fake_processor_id, fill_when_below=100500, chunk=100)

        loaded_pointer = await operations.get_or_create_pointer(fake_processor_id)

        assert loaded_pointer == ProcessorPointer(
            processor_id=fake_processor_id,
            pointer_created_at=entries_list[-1].cataloged_at,
            pointer_entry_id=entries_list[-1].id,
        )

    @pytest.mark.asyncio
    async def test_move_pointer_to_not_the_end(self, loaded_feed: Feed, fake_processor_id: int) -> None:
        await make.end_processor_pointer(1)

        entries = await l_make.n_entries(loaded_feed, 3)
        entries_list = list(entries.values())
        entries_list.sort(key=lambda entry: (entry.cataloged_at, entry.id))

        async with TableSizeDelta("ln_processors_queue", delta=2):
            await plan_processor_queue(fake_processor_id, fill_when_below=100500, chunk=2)

        loaded_pointer = await operations.get_or_create_pointer(fake_processor_id)

        assert loaded_pointer == ProcessorPointer(
            processor_id=fake_processor_id,
            pointer_created_at=entries_list[-2].cataloged_at,
            pointer_entry_id=entries_list[-2].id,
        )

    @pytest.mark.asyncio
    async def test_chunk_limit(self, loaded_feed: Feed, fake_processor_id: int) -> None:
        await operations.clear_processor_queue(fake_processor_id)

        await make.end_processor_pointer(fake_processor_id)

        entries = await l_make.n_entries(loaded_feed, 5)
        entries_list = list(entries.values())
        entries_list.sort(key=lambda entry: (entry.cataloged_at, entry.id))

        async with TableSizeDelta("ln_processors_queue", delta=3):
            await plan_processor_queue(fake_processor_id, fill_when_below=100500, chunk=3)

        loaded_pointer = await operations.get_or_create_pointer(fake_processor_id)

        assert loaded_pointer == ProcessorPointer(
            processor_id=fake_processor_id,
            pointer_created_at=entries_list[2].cataloged_at,
            pointer_entry_id=entries_list[2].id,
        )

    @pytest.mark.asyncio
    async def test_do_not_push_if_there_are_enough_entries(self, loaded_feed: Feed, fake_processor_id: int) -> None:
        await operations.clear_processor_queue(fake_processor_id)

        await make.end_processor_pointer(fake_processor_id)

        entries = await l_make.n_entries(loaded_feed, 5)
        entries_list = list(entries.values())
        entries_list.sort(key=lambda entry: (entry.cataloged_at, entry.id))

        await plan_processor_queue(fake_processor_id, fill_when_below=100500, chunk=3)

        async with TableSizeNotChanged("ln_processors_queue"):
            await plan_processor_queue(fake_processor_id, fill_when_below=3, chunk=2)

        loaded_pointer = await operations.get_or_create_pointer(fake_processor_id)

        assert loaded_pointer == ProcessorPointer(
            processor_id=fake_processor_id,
            pointer_created_at=entries_list[2].cataloged_at,
            pointer_entry_id=entries_list[2].id,
        )


class TestProcessEntry:
    @pytest.mark.asyncio
    async def test_success(
        self, fake_processor_id: int, cataloged_entry: Entry, another_cataloged_entry: Entry
    ) -> None:
        await operations.clear_processor_queue(fake_processor_id)

        await operations.push_entries_to_processor_queue(
            execute, processor_id=fake_processor_id, entry_ids=[cataloged_entry.id, another_cataloged_entry.id]
        )

        with capture_logs() as logs:
            await process_entry(
                processor_id=fake_processor_id,
                processor=AlwaysConstantProcessor(name="fake-processor", tags=["tag-1", "tag-2"]),
                entry=cataloged_entry,
            )

        assert_logs(logs, processor_successed=1, processor_requested_to_skip_entry=0, entry_processed=1)

        tags = await o_domain.get_tags_ids_for_entries([cataloged_entry.id])

        expected_ids = await o_domain.get_ids_by_uids({"tag-1", "tag-2"})  # type: ignore

        assert tags[cataloged_entry.id] == set(expected_ids.values())

        entries_in_queue = await operations.get_entries_to_process(processor_id=fake_processor_id, limit=100500)

        assert set(entries_in_queue) == {another_cataloged_entry.id}

        failed_entry_ids = await operations.get_failed_entries(execute, fake_processor_id, limit=100500)
        assert cataloged_entry.id not in failed_entry_ids

    @pytest.mark.asyncio
    async def test_skip_processing(
        self, fake_processor_id: int, cataloged_entry: Entry, another_cataloged_entry: Entry
    ) -> None:
        await operations.clear_processor_queue(fake_processor_id)

        await operations.push_entries_to_processor_queue(
            execute, processor_id=fake_processor_id, entry_ids=[cataloged_entry.id, another_cataloged_entry.id]
        )

        with capture_logs() as logs:
            await process_entry(
                processor_id=fake_processor_id,
                processor=AlwaysSkipEntryProcessor(name="fake-processor"),
                entry=cataloged_entry,
            )

        assert_logs(
            logs,
            processor_successed=0,
            processor_requested_to_skip_entry=1,
            entry_processed=1,
            processor_temporary_error=0,
        )

        tags = await o_domain.get_tags_ids_for_entries([cataloged_entry.id])

        assert cataloged_entry.id not in tags

        entries_in_queue = await operations.get_entries_to_process(processor_id=fake_processor_id, limit=100500)

        assert set(entries_in_queue) == {another_cataloged_entry.id}

        failed_entry_ids = await operations.get_failed_entries(execute, fake_processor_id, limit=100500)
        assert cataloged_entry.id not in failed_entry_ids

    @pytest.mark.asyncio
    async def test_temporary_error_in_processor(
        self, fake_processor_id: int, cataloged_entry: Entry, another_cataloged_entry: Entry
    ) -> None:
        await operations.clear_processor_queue(fake_processor_id)

        await operations.push_entries_to_processor_queue(
            execute, processor_id=fake_processor_id, entry_ids=[cataloged_entry.id, another_cataloged_entry.id]
        )

        with capture_logs() as logs:
            await process_entry(
                processor_id=fake_processor_id,
                processor=AlwaysTemporaryErrorProcessor(name="fake-processor"),
                entry=cataloged_entry,
            )

        assert_logs(
            logs,
            processor_successed=0,
            processor_requested_to_skip_entry=0,
            entry_processed=1,
            processor_temporary_error=1,
        )

        tags = await o_domain.get_tags_ids_for_entries([cataloged_entry.id])

        assert cataloged_entry.id not in tags

        entries_in_queue = await operations.get_entries_to_process(processor_id=fake_processor_id, limit=100500)

        assert set(entries_in_queue) == {another_cataloged_entry.id}

        failed_entry_ids = await operations.get_failed_entries(execute, fake_processor_id, limit=100500)
        assert cataloged_entry.id in failed_entry_ids

    @pytest.mark.asyncio
    async def test_unexpected_error(
        self, fake_processor_id: int, cataloged_entry: Entry, another_cataloged_entry: Entry
    ) -> None:
        await operations.clear_processor_queue(fake_processor_id)

        await operations.push_entries_to_processor_queue(
            execute, processor_id=fake_processor_id, entry_ids=[cataloged_entry.id, another_cataloged_entry.id]
        )

        with capture_logs() as logs:
            with pytest.raises(errors.UnexpectedErrorInProcessor):
                await process_entry(
                    processor_id=fake_processor_id,
                    processor=AlwaysErrorProcessor(name="fake-processor"),
                    entry=cataloged_entry,
                )

        assert_logs(
            logs,
            processor_successed=0,
            processor_requested_to_skip_entry=0,
            entry_processed=0,
            processor_temporary_error=0,
        )

        tags = await o_domain.get_tags_ids_for_entries([cataloged_entry.id])

        assert cataloged_entry.id not in tags

        entries_in_queue = await operations.get_entries_to_process(processor_id=fake_processor_id, limit=100500)

        assert set(entries_in_queue) == {another_cataloged_entry.id}

        failed_entry_ids = await operations.get_failed_entries(execute, fake_processor_id, limit=100500)
        assert cataloged_entry.id in failed_entry_ids


class TestMoveFailedEntriesToProcessorQueue:
    @pytest.mark.asyncio
    async def test_moved(
        self,
        fake_processor_id: int,
        another_fake_processor_id: int,
        cataloged_entry: Entry,
        another_cataloged_entry: Entry,
    ) -> None:
        await helpers.clean_failed_storage([fake_processor_id, another_fake_processor_id])
        await operations.clear_processor_queue(fake_processor_id)
        await operations.clear_processor_queue(another_fake_processor_id)

        await operations.add_entries_to_failed_storage(
            fake_processor_id, entry_ids=[cataloged_entry.id, another_cataloged_entry.id]
        )
        await operations.add_entries_to_failed_storage(
            another_fake_processor_id, entry_ids=[another_cataloged_entry.id]
        )

        async with TableSizeDelta("ln_processors_queue", delta=2):
            await move_failed_entries_to_processor_queue(fake_processor_id, limit=100500)

        entries_in_queue = await operations.get_entries_to_process(processor_id=fake_processor_id, limit=100500)

        assert set(entries_in_queue) == {cataloged_entry.id, another_cataloged_entry.id}

        async with TableSizeDelta("ln_processors_queue", delta=1):
            await move_failed_entries_to_processor_queue(another_fake_processor_id, limit=100500)

        entries_in_queue = await operations.get_entries_to_process(
            processor_id=another_fake_processor_id, limit=100500
        )

        assert set(entries_in_queue) == {another_cataloged_entry.id}

    @pytest.mark.asyncio
    async def test_limit(self, fake_processor_id: int, cataloged_entry: Entry, another_cataloged_entry: Entry) -> None:
        await helpers.clean_failed_storage([fake_processor_id])
        await operations.clear_processor_queue(fake_processor_id)

        await operations.add_entries_to_failed_storage(
            fake_processor_id, entry_ids=[cataloged_entry.id, another_cataloged_entry.id]
        )

        async with TableSizeDelta("ln_processors_queue", delta=1):
            await move_failed_entries_to_processor_queue(fake_processor_id, limit=1)

        entries_in_queue = await operations.get_entries_to_process(processor_id=fake_processor_id, limit=100500)

        assert set(entries_in_queue) == {cataloged_entry.id}



from typing import Iterable

from ffun.core import logging
from ffun.core.postgresql import ExecuteType, execute, run_in_transaction
from ffun.domain.entities import EntryId
from ffun.librarian import errors, operations
from ffun.librarian.entities import ProcessorPointer
from ffun.librarian.processors.base import Processor
from ffun.library import domain as l_domain
from ffun.library.entities import Entry
from ffun.ontology import domain as o_domain

logger = logging.get_module_logger()


count_failed_entries = operations.count_failed_entries
get_all_pointers = operations.get_all_pointers


@run_in_transaction
async def push_entries_and_move_pointer(
    execute: ExecuteType, next_pointer: ProcessorPointer, entry_ids: Iterable[EntryId]
) -> None:
    await operations.push_entries_to_processor_queue(
        execute, processor_id=next_pointer.processor_id, entry_ids=entry_ids
    )

    await operations.save_pointer(execute, pointer=next_pointer)


@run_in_transaction
async def move_failed_entries_to_processor_queue(execute: ExecuteType, processor_id: int, limit: int) -> None:
    failed_entries = await operations.get_failed_entries(execute, processor_id, limit)

    if not failed_entries:
        return

    await operations.push_entries_to_processor_queue(execute, processor_id, failed_entries)

    await operations.remove_failed_entries(execute, processor_id, failed_entries)


# most likely, this code should be in a separate worker with a more complex logic
# but for now it is ok to place it here
async def plan_processor_queue(processor_id: int, fill_when_below: int, chunk: int) -> None:
    # TODO: count may load the DB too much, because of how it is work in PostgreSQL
    #       maybe pass fill_when_below as a parameter to the function and make check on the DB side
    entries_in_queue = await operations.count_entries_in_processor_queue(processor_id)

    if entries_in_queue >= fill_when_below:
        return

    pointer = await operations.get_or_create_pointer(processor_id=processor_id)

    next_entries = await l_domain.get_entries_after_pointer(
        created_at=pointer.pointer_created_at, entry_id=pointer.pointer_entry_id, limit=chunk
    )

    if not next_entries:
        return

    last_entry = next_entries[-1]

    next_pointer = ProcessorPointer(
        processor_id=processor_id, pointer_created_at=last_entry[1], pointer_entry_id=last_entry[0]
    )

    await push_entries_and_move_pointer(next_pointer, [entry[0] for entry in next_entries])


@logging.function_args_to_log("processor.name", "entry.id")
async def process_entry(processor_id: int, processor: Processor, entry: Entry) -> None:
    logger.info("dicover_tags")

    try:
        tags = await processor.process(entry)

        tags_for_log = [tag.raw_uid for tag in tags]
        tags_for_log.sort()

        logger.info("tags_found", tags=tags_for_log)

        await o_domain.apply_tags_to_entry(entry.id, processor_id, tags)

        logger.info("processor_successed")
    except errors.SkipEntryProcessing as e:
        # do nothing in such case, see: https://github.com/Tiendil/feeds.fun/issues/176
        logger.warning("processor_requested_to_skip_entry", error_info=str(e))
    except errors.TemporaryErrorInProcessor as e:
        # log the error and move the entry to the failed storage
        # Note: it is a general plug, for some custome cases we may want to add custom processing
        # Note: currently, there are no logic to process failed storage, entries will just accumulate there
        logger.info("processor_temporary_error", error_info=str(e))
        await operations.add_entries_to_failed_storage(processor_id, [entry.id])
    except Exception as e:
        logger.exception("processor_failed")
        await operations.add_entries_to_failed_storage(processor_id, [entry.id])
        raise errors.UnexpectedErrorInProcessor(processor_id=processor_id, entry_id=entry.id) from e
    finally:
        await operations.remove_entries_from_processor_queue(execute, processor_id, [entry.id])

    logger.info("entry_processed")


@run_in_transaction
async def remove_entries_from_processor_queue(
    execute: ExecuteType, processor_id: int, entry_ids: Iterable[EntryId]
) -> None:
    await operations.remove_entries_from_processor_queue(execute, processor_id, entry_ids)



from typing import Iterable

from ffun.core.postgresql import ExecuteType, run_in_transaction
from ffun.domain.entities import UserId
from ffun.feeds.entities import FeedId
from ffun.feeds_links import operations

add_link = operations.add_link
remove_link = operations.remove_link
get_link = operations.get_link
get_linked_feeds = operations.get_linked_feeds
get_linked_users = operations.get_linked_users
has_linked_users = operations.has_linked_users
tech_remove_all_links = operations.tech_remove_all_links
count_feeds_per_user = operations.count_feeds_per_user
count_subset_feeds_per_user = operations.count_subset_feeds_per_user


async def get_linked_users_flat(feed_ids: Iterable[FeedId]) -> set[UserId]:
    users = await operations.get_linked_users(feed_ids)

    answer = set()

    for user_set in users.values():
        answer.update(user_set)

    return answer


@run_in_transaction
async def tech_merge_feeds(execute: ExecuteType, from_feed_id: FeedId, to_feed_id: FeedId) -> None:
    await operations.tech_merge_feeds(execute, from_feed_id=from_feed_id, to_feed_id=to_feed_id)



import asyncio
import contextlib

from ffun.application import utils as app_utils
from ffun.application import workers as app_workers
from ffun.application.application import with_app
from ffun.cli.application import app
from ffun.loader import domain as l_domain


async def run(loader: bool, librarian: bool) -> None:
    async with with_app():
        async with contextlib.AsyncExitStack() as stack:
            if loader:
                l_domain.initialize(user_agent=app_utils.user_agent())
                await stack.enter_async_context(app_workers.use_loader())

            if librarian:
                await stack.enter_async_context(app_workers.use_librarian())

            while True:
                await asyncio.sleep(0.1)


@app.command()
def workers(loader: bool = False, librarian: bool = False) -> None:
    asyncio.run(run(loader=loader, librarian=librarian))



from typing import Iterable

from ffun.core.postgresql import execute
from ffun.domain.entities import EntryId
from ffun.ontology.operations import get_tags_for_entries


async def assert_has_tags(tags_ids: dict[EntryId, Iterable[int]]) -> None:
    tags = await get_tags_for_entries(execute, [entry_id for entry_id in tags_ids])

    for entry_id, tag_ids in tags_ids.items():
        assert tags.get(entry_id, set()) >= set(tag_ids)
